# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import glance_store

from glance.common import store_utils
import glance.db
from glance.glare import dependency
from glance.glare import domain
from glance.glare import location
from glance.glare import updater


class Gateway(object):
    def __init__(self, db_api=None, store_api=None, plugins=None):
        self.db_api = db_api or glance.db.get_api()
        self.store_api = store_api or glance_store
        self.store_utils = store_utils
        self.plugins = plugins

    def get_artifact_type_factory(self, context, klass):
        declarative_factory = domain.ArtifactFactory(context, klass)
        repo = self.get_artifact_repo(context)
        dependencies_factory = dependency.ArtifactFactory(declarative_factory,
                                                          klass, repo)
        factory = location.ArtifactFactoryProxy(dependencies_factory,
                                                context,
                                                self.store_api,
                                                self.store_utils)
        updater_factory = updater.ArtifactFactoryProxy(factory)
        return updater_factory

    def get_artifact_repo(self, context):
        artifact_repo = glance.db.ArtifactRepo(context,
                                               self.db_api,
                                               self.plugins)
        dependencies_repo = dependency.ArtifactRepo(artifact_repo,
                                                    self.plugins)
        repo = location.ArtifactRepoProxy(dependencies_repo,
                                          context,
                                          self.store_api,
                                          self.store_utils)
        updater_repo = updater.ArtifactRepoProxy(repo)
        return updater_repo



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import glance_store as store
from glance_store import location
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_db import options
from oslo_serialization import jsonutils

from glance.tests import stubs
from glance.tests import utils as test_utils

CONF = cfg.CONF


class StoreClearingUnitTest(test_utils.BaseTestCase):

    def setUp(self):
        super(StoreClearingUnitTest, self).setUp()
        # Ensure stores + locations cleared
        location.SCHEME_TO_CLS_MAP = {}

        self._create_stores()
        self.addCleanup(setattr, location, 'SCHEME_TO_CLS_MAP', dict())

    def _create_stores(self, passing_config=True):
        """Create known stores. Mock out sheepdog's subprocess dependency
        on collie.

        :param passing_config: making store driver passes basic configurations.
        :returns: the number of how many store drivers been loaded.
        """
        store.register_opts(CONF)

        self.config(default_store='filesystem',
                    filesystem_store_datadir=self.test_dir,
                    group="glance_store")

        store.create_stores(CONF)


class IsolatedUnitTest(StoreClearingUnitTest):

    """
    Unit test case that establishes a mock environment within
    a testing directory (in isolation)
    """
    registry = None

    def setUp(self):
        super(IsolatedUnitTest, self).setUp()
        options.set_defaults(CONF, connection='sqlite://')
        lockutils.set_defaults(os.path.join(self.test_dir))

        self.config(debug=False)

        self.config(default_store='filesystem',
                    filesystem_store_datadir=self.test_dir,
                    group="glance_store")

        store.create_stores()
        stubs.stub_out_registry_and_store_server(self.stubs,
                                                 self.test_dir,
                                                 registry=self.registry)

    def set_policy_rules(self, rules):
        fap = open(CONF.oslo_policy.policy_file, 'w')
        fap.write(jsonutils.dumps(rules))
        fap.close()



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import operator
import uuid

from enum import Enum
from oslo_config import cfg
from oslo_db import exception as db_exc
import sqlalchemy
from sqlalchemy import and_
from sqlalchemy import case
from sqlalchemy import or_
import sqlalchemy.orm as orm
from sqlalchemy.orm import joinedload

from glance.common import exception
from glance.common import semver_db
from glance.common import timeutils
from glance.db.sqlalchemy import models_glare as models
import glance.glare as ga
from glance.i18n import _LE, _LW
from oslo_log import log as os_logging

LOG = os_logging.getLogger(__name__)

CONF = cfg.CONF


class Visibility(Enum):
    PRIVATE = 'private'
    PUBLIC = 'public'
    SHARED = 'shared'


class State(Enum):
    CREATING = 'creating'
    ACTIVE = 'active'
    DEACTIVATED = 'deactivated'
    DELETED = 'deleted'


TRANSITIONS = {
    State.CREATING: [State.ACTIVE, State.DELETED],
    State.ACTIVE: [State.DEACTIVATED, State.DELETED],
    State.DEACTIVATED: [State.ACTIVE, State.DELETED],
    State.DELETED: []
}


def create(context, values, session, type_name, type_version=None):
    return _out(_create_or_update(context, values, None, session,
                                  type_name, type_version))


def update(context, values, artifact_id, session,
           type_name, type_version=None):
    return _out(_create_or_update(context, values, artifact_id, session,
                                  type_name, type_version))


def delete(context, artifact_id, session, type_name, type_version=None):
    values = {'state': 'deleted'}
    return _out(_create_or_update(context, values, artifact_id, session,
                                  type_name, type_version))


def _create_or_update(context, values, artifact_id, session, type_name,
                      type_version=None):
    values = copy.deepcopy(values)
    with session.begin():
        _set_version_fields(values)
        _validate_values(values)
        _drop_protected_attrs(models.Artifact, values)
        if artifact_id:
            # update existing artifact
            state = values.get('state')
            show_level = ga.Showlevel.BASIC
            if state is not None:
                if state == 'active':
                    show_level = ga.Showlevel.DIRECT
                    values['published_at'] = timeutils.utcnow()
                if state == 'deleted':
                    values['deleted_at'] = timeutils.utcnow()

            artifact = _get(context, artifact_id, session, type_name,
                            type_version, show_level=show_level)
            _validate_transition(artifact.state,
                                 values.get('state') or artifact.state)
        else:
            # create new artifact
            artifact = models.Artifact()
            if 'id' not in values:
                artifact.id = str(uuid.uuid4())
            else:
                artifact.id = values['id']

        if 'tags' in values:
            tags = values.pop('tags')
            artifact.tags = _do_tags(artifact, tags)

        if 'properties' in values:
            properties = values.pop('properties', {})
            artifact.properties = _do_properties(artifact, properties)

        if 'blobs' in values:
            blobs = values.pop('blobs')
            artifact.blobs = _do_blobs(artifact, blobs)

        if 'dependencies' in values:
            dependencies = values.pop('dependencies')
            _do_dependencies(artifact, dependencies, session)

        if values.get('state', None) == 'publish':
            artifact.dependencies.extend(
                _do_transitive_dependencies(artifact, session))

        artifact.update(values)
        try:
            artifact.save(session=session)
        except db_exc.DBDuplicateEntry:
            LOG.warn(_LW("Artifact with the specified type, name and version "
                         "already exists"))
            raise exception.ArtifactDuplicateNameTypeVersion()

    return artifact


def get(context, artifact_id, session, type_name=None, type_version=None,
        show_level=ga.Showlevel.BASIC):
    artifact = _get(context, artifact_id, session, type_name, type_version,
                    show_level)
    return _out(artifact, show_level)


def publish(context, artifact_id, session, type_name,
            type_version=None):
    """
    Because transitive dependencies are not initially created it has to be done
    manually by calling this function.
    It creates transitive dependencies for the given artifact_id and saves
    them in DB.
    :returns: artifact dict with Transitive show level
    """
    values = {'state': 'active'}
    return _out(_create_or_update(context, values, artifact_id, session,
                                  type_name, type_version))


def _validate_transition(source_state, target_state):
    if target_state == source_state:
        return
    try:
        source_state = State(source_state)
        target_state = State(target_state)
    except ValueError:
        raise exception.InvalidArtifactStateTransition(source=source_state,
                                                       target=target_state)
    if (source_state not in TRANSITIONS or
       target_state not in TRANSITIONS[source_state]):
        raise exception.InvalidArtifactStateTransition(source=source_state,
                                                       target=target_state)


def _out(artifact, show_level=ga.Showlevel.BASIC, show_text_properties=True):
    """
    Transforms sqlalchemy object into dict depending on the show level.

    :param artifact: sql
    :param show_level: constant from Showlevel class
    :param show_text_properties: for performance optimization it's possible
    to disable loading of massive text properties
    :returns: generated dict
    """
    res = artifact.to_dict(show_level=show_level,
                           show_text_properties=show_text_properties)

    if show_level >= ga.Showlevel.DIRECT:
        dependencies = artifact.dependencies
        dependencies.sort(key=lambda elem: (elem.artifact_origin,
                                            elem.name, elem.position))
        res['dependencies'] = {}
        if show_level == ga.Showlevel.DIRECT:
            new_show_level = ga.Showlevel.BASIC
        else:
            new_show_level = ga.Showlevel.TRANSITIVE
        for dep in dependencies:
            if dep.artifact_origin == artifact.id:
                # make array
                for p in res['dependencies'].keys():
                    if p == dep.name:
                        # add value to array
                        res['dependencies'][p].append(
                            _out(dep.dest, new_show_level))
                        break
                else:
                    # create new array
                    deparr = [_out(dep.dest, new_show_level)]
                    res['dependencies'][dep.name] = deparr
    return res


def _get(context, artifact_id, session, type_name=None, type_version=None,
         show_level=ga.Showlevel.BASIC):
    values = dict(id=artifact_id)
    if type_name is not None:
        values['type_name'] = type_name
    if type_version is not None:
        values['type_version'] = type_version
    _set_version_fields(values)
    try:
        if show_level == ga.Showlevel.NONE:
            query = (
                session.query(models.Artifact).
                options(joinedload(models.Artifact.tags)).
                filter_by(**values))
        else:
            query = (
                session.query(models.Artifact).
                options(joinedload(models.Artifact.properties)).
                options(joinedload(models.Artifact.tags)).
                options(joinedload(models.Artifact.blobs).
                        joinedload(models.ArtifactBlob.locations)).
                filter_by(**values))

        artifact = query.one()
    except orm.exc.NoResultFound:
        LOG.warn(_LW("Artifact with id=%s not found") % artifact_id)
        raise exception.ArtifactNotFound(id=artifact_id)
    if not _check_visibility(context, artifact):
        LOG.warn(_LW("Artifact with id=%s is not accessible") % artifact_id)
        raise exception.ArtifactForbidden(id=artifact_id)
    return artifact


def get_all(context, session, marker=None, limit=None,
            sort_keys=None, sort_dirs=None, filters=None,
            show_level=ga.Showlevel.NONE):
    """List all visible artifacts"""

    filters = filters or {}

    artifacts = _get_all(
        context, session, filters, marker,
        limit, sort_keys, sort_dirs, show_level)

    return map(lambda ns: _out(ns, show_level, show_text_properties=False),
               artifacts)


def _get_all(context, session, filters=None, marker=None,
             limit=None, sort_keys=None, sort_dirs=None,
             show_level=ga.Showlevel.NONE):
    """Get all namespaces that match zero or more filters.

    :param filters: dict of filter keys and values.
    :param marker: namespace id after which to start page
    :param limit: maximum number of namespaces to return
    :param sort_keys: namespace attributes by which results should be sorted
    :param sort_dirs: directions in which results should be sorted (asc, desc)
    """

    filters = filters or {}

    query = _do_artifacts_query(context, session, show_level)
    basic_conds, tag_conds, prop_conds = _do_query_filters(filters)

    if basic_conds:
        for basic_condition in basic_conds:
            query = query.filter(and_(*basic_condition))

    if tag_conds:
        for tag_condition in tag_conds:
            query = query.join(models.ArtifactTag, aliased=True).filter(
                and_(*tag_condition))

    if prop_conds:
        for prop_condition in prop_conds:
            query = query.join(models.ArtifactProperty, aliased=True).filter(
                and_(*prop_condition))

    marker_artifact = None
    if marker is not None:
        marker_artifact = _get(context, marker, session, None, None)

    if sort_keys is None:
        sort_keys = [('created_at', None), ('id', None)]
        sort_dirs = ['desc', 'desc']
    else:
        for key in [('created_at', None), ('id', None)]:
            if key not in sort_keys:
                sort_keys.append(key)
                sort_dirs.append('desc')

    # Note(mfedosin): Workaround to deal with situation that sqlalchemy cannot
    # work with composite keys correctly
    if ('version', None) in sort_keys:
        i = sort_keys.index(('version', None))
        version_sort_dir = sort_dirs[i]
        sort_keys[i:i + 1] = [('version_prefix', None),
                              ('version_suffix', None),
                              ('version_meta', None)]
        sort_dirs[i:i + 1] = [version_sort_dir] * 3

    query = _do_paginate_query(query=query,
                               limit=limit,
                               sort_keys=sort_keys,
                               marker=marker_artifact,
                               sort_dirs=sort_dirs)

    return query.all()


def _do_paginate_query(query, sort_keys=None, sort_dirs=None,
                       marker=None, limit=None):
    # Default the sort direction to ascending
    sort_dir = 'asc'

    # Ensure a per-column sort direction
    if sort_dirs is None:
        sort_dirs = [sort_dir] * len(sort_keys)

    assert(len(sort_dirs) == len(sort_keys))  # nosec
    # nosec: This function runs safely if the assertion fails.
    if len(sort_dirs) < len(sort_keys):
        sort_dirs += [sort_dir] * (len(sort_keys) - len(sort_dirs))

    # Add sorting
    for current_sort_key, current_sort_dir in zip(sort_keys, sort_dirs):
        try:
            sort_dir_func = {
                'asc': sqlalchemy.asc,
                'desc': sqlalchemy.desc,
            }[current_sort_dir]
        except KeyError:
            raise ValueError(_LE("Unknown sort direction, "
                                 "must be 'desc' or 'asc'"))

        if current_sort_key[1] is None:
            # sort by generic property
            query = query.order_by(sort_dir_func(getattr(
                models.Artifact,
                current_sort_key[0])))
        else:
            # sort by custom property
            prop_type = current_sort_key[1] + "_value"
            query = (
                query.join(models.ArtifactProperty).
                filter(models.ArtifactProperty.name == current_sort_key[0]).
                order_by(sort_dir_func(getattr(models.ArtifactProperty,
                                               prop_type))))

    default = ''

    # Add pagination
    if marker is not None:
        marker_values = []
        for sort_key in sort_keys:
            v = getattr(marker, sort_key[0])
            if v is None:
                v = default
            marker_values.append(v)

        # Build up an array of sort criteria as in the docstring
        criteria_list = []
        for i in range(len(sort_keys)):
            crit_attrs = []
            if marker_values[i] is None:
                continue
            for j in range(i):
                if sort_keys[j][1] is None:
                    model_attr = getattr(models.Artifact, sort_keys[j][0])
                else:
                    model_attr = getattr(models.ArtifactProperty,
                                         sort_keys[j][1] + "_value")
                default = None if isinstance(
                    model_attr.property.columns[0].type,
                    sqlalchemy.DateTime) else ''
                attr = case([(model_attr != None,
                              model_attr), ],
                            else_=default)
                crit_attrs.append((attr == marker_values[j]))

            if sort_keys[i][1] is None:
                model_attr = getattr(models.Artifact, sort_keys[i][0])
            else:
                model_attr = getattr(models.ArtifactProperty,
                                     sort_keys[i][1] + "_value")

            default = None if isinstance(model_attr.property.columns[0].type,
                                         sqlalchemy.DateTime) else ''
            attr = case([(model_attr != None,
                          model_attr), ],
                        else_=default)

            if sort_dirs[i] == 'desc':
                crit_attrs.append((attr < marker_values[i]))
            else:
                crit_attrs.append((attr > marker_values[i]))

            criteria = and_(*crit_attrs)
            criteria_list.append(criteria)

        f = or_(*criteria_list)
        query = query.filter(f)

    if limit is not None:
        query = query.limit(limit)

    return query


def _do_artifacts_query(context, session, show_level=ga.Showlevel.NONE):
    """Build the query to get all artifacts based on the context"""

    LOG.debug("context.is_admin=%(is_admin)s; context.owner=%(owner)s",
              {'is_admin': context.is_admin, 'owner': context.owner})

    if show_level == ga.Showlevel.NONE:
        query = session.query(models.Artifact).options(
            joinedload(models.Artifact.tags))
    elif show_level == ga.Showlevel.BASIC:
        query = (
            session.query(models.Artifact).
            options(joinedload(
                models.Artifact.properties).
                defer(models.ArtifactProperty.text_value)).
            options(joinedload(models.Artifact.tags)).
            options(joinedload(models.Artifact.blobs).
                    joinedload(models.ArtifactBlob.locations)))
    else:
        # other show_levels aren't supported
        msg = _LW("Show level %s is not supported in this "
                  "operation") % ga.Showlevel.to_str(show_level)
        LOG.warn(msg)
        raise exception.ArtifactUnsupportedShowLevel(shl=show_level)

    # If admin, return everything.
    if context.is_admin:
        return query
    else:
        # If regular user, return only public artifacts.
        # However, if context.owner has a value, return both
        # public and private artifacts of the context.owner.
        if context.owner is not None:
            query = query.filter(
                or_(models.Artifact.owner == context.owner,
                    models.Artifact.visibility == 'public'))
        else:
            query = query.filter(
                models.Artifact.visibility == 'public')
        return query

op_mappings = {
    'EQ': operator.eq,
    'GT': operator.gt,
    'GE': operator.ge,
    'LT': operator.lt,
    'LE': operator.le,
    'NE': operator.ne,
    'IN': operator.eq  # it must be eq
}


def _do_query_filters(filters):
    basic_conds = []
    tag_conds = []
    prop_conds = []

    # don't show deleted artifacts
    basic_conds.append([models.Artifact.state != 'deleted'])

    visibility = filters.pop('visibility', None)
    if visibility is not None:
        # ignore operator. always consider it EQ
        basic_conds.append(
            [models.Artifact.visibility == visibility[0]['value']])

    type_name = filters.pop('type_name', None)
    if type_name is not None:
        # ignore operator. always consider it EQ
        basic_conds.append([models.Artifact.type_name == type_name['value']])
        type_version = filters.pop('type_version', None)
        if type_version is not None:
            # ignore operator. always consider it EQ
            # TODO(mfedosin) add support of LIKE operator
            type_version = semver_db.parse(type_version['value'])
            basic_conds.append([models.Artifact.type_version == type_version])

    name = filters.pop('name', None)
    if name is not None:
        # ignore operator. always consider it EQ
        basic_conds.append([models.Artifact.name == name[0]['value']])

    versions = filters.pop('version', None)
    if versions is not None:
        for version in versions:
            value = semver_db.parse(version['value'])
            op = version['operator']
            fn = op_mappings[op]
            basic_conds.append([fn(models.Artifact.version, value)])

    state = filters.pop('state', None)
    if state is not None:
        # ignore operator. always consider it EQ
        basic_conds.append([models.Artifact.state == state['value']])

    owner = filters.pop('owner', None)
    if owner is not None:
        # ignore operator. always consider it EQ
        basic_conds.append([models.Artifact.owner == owner[0]['value']])

    id_list = filters.pop('id_list', None)
    if id_list is not None:
        basic_conds.append([models.Artifact.id.in_(id_list['value'])])

    name_list = filters.pop('name_list', None)
    if name_list is not None:
        basic_conds.append([models.Artifact.name.in_(name_list['value'])])

    tags = filters.pop('tags', None)
    if tags is not None:
        for tag in tags:
            tag_conds.append([models.ArtifactTag.value == tag['value']])

    # process remaining filters
    for filtername, filtervalues in filters.items():
        for filtervalue in filtervalues:

            db_prop_op = filtervalue['operator']
            db_prop_value = filtervalue['value']
            db_prop_type = filtervalue['type'] + "_value"
            db_prop_position = filtervalue.get('position')

            conds = [models.ArtifactProperty.name == filtername]

            if db_prop_op in op_mappings:
                fn = op_mappings[db_prop_op]
                result = fn(getattr(models.ArtifactProperty, db_prop_type),
                            db_prop_value)

                cond = [result]
                if db_prop_position is not 'any':
                    cond.append(
                        models.ArtifactProperty.position == db_prop_position)
                if db_prop_op == 'IN':
                    if (db_prop_position is not None and
                            db_prop_position is not 'any'):
                        msg = _LE("Cannot use this parameter with "
                                  "the operator IN")
                        LOG.error(msg)
                        raise exception.ArtifactInvalidPropertyParameter(
                            op='IN')
                    cond = [result,
                            models.ArtifactProperty.position >= 0]
            else:
                msg = _LE("Operator %s is not supported") % db_prop_op
                LOG.error(msg)
                raise exception.ArtifactUnsupportedPropertyOperator(
                    op=db_prop_op)

            conds.extend(cond)

            prop_conds.append(conds)
    return basic_conds, tag_conds, prop_conds


def _do_tags(artifact, new_tags):
    tags_to_update = []
    # don't touch existing tags
    for tag in artifact.tags:
        if tag.value in new_tags:
            tags_to_update.append(tag)
            new_tags.remove(tag.value)
    # add new tags
    for tag in new_tags:
        db_tag = models.ArtifactTag()
        db_tag.value = tag
        tags_to_update.append(db_tag)
    return tags_to_update


def _do_property(propname, prop, position=None):
    db_prop = models.ArtifactProperty()
    db_prop.name = propname
    setattr(db_prop,
            (prop['type'] + "_value"),
            prop['value'])
    db_prop.position = position
    return db_prop


def _do_properties(artifact, new_properties):

    props_to_update = []
    # don't touch existing properties
    for prop in artifact.properties:
        if prop.name not in new_properties:
            props_to_update.append(prop)

    for propname, prop in new_properties.items():
        if prop['type'] == 'array':
            for pos, arrprop in enumerate(prop['value']):
                props_to_update.append(
                    _do_property(propname, arrprop, pos)
                )
        else:
            props_to_update.append(
                _do_property(propname, prop)
            )
    return props_to_update


def _do_blobs(artifact, new_blobs):
    blobs_to_update = []

    # don't touch existing blobs
    for blob in artifact.blobs:
        if blob.name not in new_blobs:
            blobs_to_update.append(blob)

    for blobname, blobs in new_blobs.items():
        for pos, blob in enumerate(blobs):
            for db_blob in artifact.blobs:
                if db_blob.name == blobname and db_blob.position == pos:
                    # update existing blobs
                    db_blob.size = blob['size']
                    db_blob.checksum = blob['checksum']
                    db_blob.item_key = blob['item_key']
                    db_blob.locations = _do_locations(db_blob,
                                                      blob['locations'])
                    blobs_to_update.append(db_blob)
                    break
            else:
                # create new blob
                db_blob = models.ArtifactBlob()
                db_blob.name = blobname
                db_blob.size = blob['size']
                db_blob.checksum = blob['checksum']
                db_blob.item_key = blob['item_key']
                db_blob.position = pos
                db_blob.locations = _do_locations(db_blob, blob['locations'])
                blobs_to_update.append(db_blob)
    return blobs_to_update


def _do_locations(blob, new_locations):
    locs_to_update = []
    for pos, loc in enumerate(new_locations):
        for db_loc in blob.locations:
            if db_loc.value == loc['value']:
                # update existing location
                db_loc.position = pos
                db_loc.status = loc['status']
                locs_to_update.append(db_loc)
                break
        else:
            # create new location
            db_loc = models.ArtifactBlobLocation()
            db_loc.value = loc['value']
            db_loc.status = loc['status']
            db_loc.position = pos
            locs_to_update.append(db_loc)
    return locs_to_update


def _do_dependencies(artifact, new_dependencies, session):
    deps_to_update = []
    # small check that all dependencies are new
    if artifact.dependencies is not None:
        for db_dep in artifact.dependencies:
            for dep in new_dependencies.keys():
                if db_dep.name == dep:
                    msg = _LW("Artifact with the specified type, name "
                              "and versions already has the direct "
                              "dependency=%s") % dep
                    LOG.warn(msg)
        # change values of former dependency
        for dep in artifact.dependencies:
            session.delete(dep)
        artifact.dependencies = []
    for depname, depvalues in new_dependencies.items():
        for pos, depvalue in enumerate(depvalues):
            db_dep = models.ArtifactDependency()
            db_dep.name = depname
            db_dep.artifact_source = artifact.id
            db_dep.artifact_dest = depvalue
            db_dep.artifact_origin = artifact.id
            db_dep.is_direct = True
            db_dep.position = pos
            deps_to_update.append(db_dep)
    artifact.dependencies = deps_to_update


def _do_transitive_dependencies(artifact, session):
    deps_to_update = []
    for dependency in artifact.dependencies:
        depvalue = dependency.artifact_dest
        transitdeps = session.query(models.ArtifactDependency).filter_by(
            artifact_source=depvalue).all()
        for transitdep in transitdeps:
            if not transitdep.is_direct:
                # transitive dependencies are already created
                msg = _LW("Artifact with the specified type, "
                          "name and version already has the "
                          "direct dependency=%d") % transitdep.id
                LOG.warn(msg)
                raise exception.ArtifactDuplicateTransitiveDependency(
                    dep=transitdep.id)

            db_dep = models.ArtifactDependency()
            db_dep.name = transitdep['name']
            db_dep.artifact_source = artifact.id
            db_dep.artifact_dest = transitdep.artifact_dest
            db_dep.artifact_origin = transitdep.artifact_source
            db_dep.is_direct = False
            db_dep.position = transitdep.position
            deps_to_update.append(db_dep)
    return deps_to_update


def _check_visibility(context, artifact):
    if context.is_admin:
        return True

    if not artifact.owner:
        return True

    if artifact.visibility == Visibility.PUBLIC.value:
        return True

    if artifact.visibility == Visibility.PRIVATE.value:
        if context.owner and context.owner == artifact.owner:
            return True
        else:
            return False

    if artifact.visibility == Visibility.SHARED.value:
        return False

    return False


def _set_version_fields(values):
    if 'type_version' in values:
        values['type_version'] = semver_db.parse(values['type_version'])
    if 'version' in values:
        values['version'] = semver_db.parse(values['version'])


def _validate_values(values):
    if 'state' in values:
        try:
            State(values['state'])
        except ValueError:
            msg = "Invalid artifact state '%s'" % values['state']
            raise exception.Invalid(msg)
    if 'visibility' in values:
        try:
            Visibility(values['visibility'])
        except ValueError:
            msg = "Invalid artifact visibility '%s'" % values['visibility']
            raise exception.Invalid(msg)
    # TODO(mfedosin): it's an idea to validate tags someday
    # (check that all tags match the regexp)


def _drop_protected_attrs(model_class, values):
    """
    Removed protected attributes from values dictionary using the models
    __protected_attributes__ field.
    """
    for attr in model_class.__protected_attributes__:
        if attr in values:
            del values[attr]



# Copyright 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
A mixin that validates the given body for jsonpatch-compatibility.
The methods supported are limited to listed in METHODS_ALLOWED
"""

import re

import jsonschema

import glance.common.exception as exc
from glance.i18n import _


class JsonPatchValidatorMixin(object):
    # a list of allowed methods allowed according to RFC 6902
    ALLOWED = ["replace", "test", "remove", "add", "copy"]
    PATH_REGEX_COMPILED = re.compile("^/[^/]+(/[^/]+)*$")

    def __init__(self, methods_allowed=None):
        if methods_allowed is None:
            methods_allowed = ["replace", "remove"]
        self.schema = self._gen_schema(methods_allowed)
        self.methods_allowed = [m for m in methods_allowed
                                if m in self.ALLOWED]

    @staticmethod
    def _gen_schema(methods_allowed):
        """
        Generates a jsonschema for jsonpatch request based on methods_allowed
        """
        # op replace needs no 'value' param, so needs a special schema if
        # present in methods_allowed
        basic_schema = {
            "type": "array",
            "items": {"properties": {"op": {"type": "string",
                                            "enum": methods_allowed},
                                     "path": {"type": "string"},
                                     "value": {"type": ["string",
                                                        "object",
                                                        "integer",
                                                        "array",
                                                        "boolean"]}
                                     },
                      "required": ["op", "path", "value"],
                      "type": "object"},
            "$schema": "http://json-schema.org/draft-04/schema#"
        }
        if "remove" in methods_allowed:
            methods_allowed.remove("remove")
            no_remove_op_schema = {
                "type": "object",
                "properties": {
                    "op": {"type": "string", "enum": methods_allowed},
                    "path": {"type": "string"},
                    "value": {"type": ["string", "object",
                                       "integer", "array", "boolean"]}
                },
                "required": ["op", "path", "value"]}
            op_remove_only_schema = {
                "type": "object",
                "properties": {
                    "op": {"type": "string", "enum": ["remove"]},
                    "path": {"type": "string"}
                },
                "required": ["op", "path"]}

            basic_schema = {
                "type": "array",
                "items": {
                    "oneOf": [no_remove_op_schema, op_remove_only_schema]},
                "$schema": "http://json-schema.org/draft-04/schema#"
            }
        return basic_schema

    def validate_body(self, body):
        try:
            jsonschema.validate(body, self.schema)
            # now make sure everything is ok with path
            return [{"path": self._decode_json_pointer(e["path"]),
                     "value": e.get("value", None),
                     "op": e["op"]} for e in body]
        except jsonschema.ValidationError:
            raise exc.InvalidJsonPatchBody(body=body, schema=self.schema)

    def _check_for_path_errors(self, pointer):
        if not re.match(self.PATH_REGEX_COMPILED, pointer):
            msg = _("Json path should start with a '/', "
                    "end with no '/', no 2 subsequent '/' are allowed.")
            raise exc.InvalidJsonPatchPath(path=pointer, explanation=msg)
        if re.search('~[^01]', pointer) or pointer.endswith('~'):
            msg = _("Pointer contains '~' which is not part of"
                    " a recognized escape sequence [~0, ~1].")
            raise exc.InvalidJsonPatchPath(path=pointer, explanation=msg)

    def _decode_json_pointer(self, pointer):
        """Parses a json pointer. Returns a pointer as a string.

        Json Pointers are defined in
        http://tools.ietf.org/html/draft-pbryan-zyp-json-pointer .
        The pointers use '/' for separation between object attributes.
        A '/' character in an attribute name is encoded as "~1" and
        a '~' character is encoded as "~0".
        """
        self._check_for_path_errors(pointer)
        ret = []
        for part in pointer.lstrip('/').split('/'):
            ret.append(part.replace('~1', '/').replace('~0', '~').strip())
        return '/'.join(ret)



# Copyright 2011-2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from oslo_config import cfg
import semantic_version
from stevedore import enabled

from glance.common import exception
from glance.common.glare import definitions
from glance.i18n import _, _LE, _LI, _LW
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

plugins_opts = [
    cfg.BoolOpt('load_enabled', default=True,
                help=_('When false, no artifacts can be loaded regardless of'
                       ' available_plugins. When true, artifacts can be'
                       ' loaded.')),
    cfg.ListOpt('available_plugins', default=[],
                help=_('A list of artifacts that are allowed in the'
                       ' format name or name-version. Empty list means that'
                       ' any artifact can be loaded.'))
]


CONF = cfg.CONF
CONF.register_opts(plugins_opts)


class ArtifactsPluginLoader(object):
    def __init__(self, namespace):
        self.mgr = enabled.EnabledExtensionManager(
            check_func=self._gen_check_func(),
            namespace=namespace,
            propagate_map_exceptions=True,
            on_load_failure_callback=self._on_load_failure)
        self.plugin_map = {'by_typename': {},
                           'by_endpoint': {}}

        def _add_extension(ext):
            """
            Plugins can be loaded as entry_point=single plugin and
            entry_point=PLUGIN_LIST, where PLUGIN_LIST is a python variable
            holding a list of plugins
            """
            def _load_one(plugin):
                if issubclass(plugin, definitions.ArtifactType):
                    # make sure that have correct plugin name
                    art_name = plugin.metadata.type_name
                    if art_name != ext.name:
                        raise exception.ArtifactNonMatchingTypeName(
                            name=art_name, plugin=ext.name)
                    # make sure that no plugin with the same name and version
                    # already exists
                    exists = self._get_plugins(ext.name)
                    new_tv = plugin.metadata.type_version
                    if any(e.metadata.type_version == new_tv for e in exists):
                        raise exception.ArtifactDuplicateNameTypeVersion()
                self._add_plugin("by_endpoint", plugin.metadata.endpoint,
                                 plugin)
                self._add_plugin("by_typename", plugin.metadata.type_name,
                                 plugin)

            if isinstance(ext.plugin, list):
                for p in ext.plugin:
                    _load_one(p)
            else:
                _load_one(ext.plugin)

        # (ivasilevskaya) that looks pretty bad as RuntimeError is too general,
        # but stevedore has awful exception wrapping with no specific class
        # for this very case (no extensions for given namespace found)
        try:
            self.mgr.map(_add_extension)
        except RuntimeError as re:
            LOG.error(_LE("Unable to load artifacts: %s") % re.message)

    def _version(self, artifact):
        return semantic_version.Version.coerce(artifact.metadata.type_version)

    def _add_plugin(self, spec, name, plugin):
        """
        Inserts a new plugin into a sorted by desc type_version list
        of existing plugins in order to retrieve the latest by next()
        """
        def _add(name, value):
            self.plugin_map[spec][name] = value

        old_order = copy.copy(self._get_plugins(name, spec=spec))
        for i, p in enumerate(old_order):
            if self._version(p) < self._version(plugin):
                _add(name, old_order[0:i] + [plugin] + old_order[i:])
                return
        _add(name, old_order + [plugin])

    def _get_plugins(self, name, spec="by_typename"):
        if spec not in self.plugin_map.keys():
            return []
        return self.plugin_map[spec].get(name, [])

    def _gen_check_func(self):
        """generates check_func for EnabledExtensionManager"""

        def _all_forbidden(ext):
            LOG.warn(_LW("Can't load artifact %s: load disabled in config") %
                     ext.name)
            raise exception.ArtifactLoadError(name=ext.name)

        def _all_allowed(ext):
            LOG.info(
                _LI("Artifact %s has been successfully loaded"), ext.name)
            return True

        if not CONF.load_enabled:
            return _all_forbidden
        if len(CONF.available_plugins) == 0:
            return _all_allowed

        available = []
        for name in CONF.available_plugins:
            type_name, version = (name.split('-', 1)
                                  if '-' in name else (name, None))
            available.append((type_name, version))

        def _check_ext(ext):
            try:
                next(n for n, v in available
                     if n == ext.plugin.metadata.type_name and
                     (v is None or v == ext.plugin.metadata.type_version))
            except StopIteration:
                LOG.warn(_LW("Can't load artifact %s: not in"
                             " available_plugins list") % ext.name)
                raise exception.ArtifactLoadError(name=ext.name)
            LOG.info(
                _LI("Artifact %s has been successfully loaded"), ext.name)
            return True

        return _check_ext

    # this has to be done explicitly as stevedore is pretty ignorant when
    # face to face with an Exception and tries to swallow it and print sth
    # irrelevant instead of expected error message
    def _on_load_failure(self, manager, ep, exc):
        msg = (_LE("Could not load plugin from %(module)s") %
               {"module": ep.module_name})
        LOG.exception(msg)
        raise exc

    def _find_class_in_collection(self, collection, name, version=None):
        try:
            def _cmp_version(plugin, version):
                ver = semantic_version.Version.coerce
                return (ver(plugin.metadata.type_version) ==
                        ver(version))

            if version:
                return next((p for p in collection
                             if _cmp_version(p, version)))
            return next((p for p in collection))
        except StopIteration:
            raise exception.ArtifactPluginNotFound(
                name="%s %s" % (name, "v %s" % version if version else ""))

    def get_class_by_endpoint(self, name, version=None):
        if version is None:
            classlist = self._get_plugins(name, spec="by_endpoint")
            if not classlist:
                raise exception.ArtifactPluginNotFound(name=name)
            return self._find_class_in_collection(classlist, name)
        return self._find_class_in_collection(
            self._get_plugins(name, spec="by_endpoint"), name, version)

    def get_class_by_typename(self, name, version=None):
        return self._find_class_in_collection(
            self._get_plugins(name, spec="by_typename"), name, version)



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import glance.common.exception as exc
import glance.common.jsonpatchvalidator as jpv
import glance.tests.utils as utils


class TestValidator(jpv.JsonPatchValidatorMixin):
    def __init__(self, methods_allowed=None):
        if methods_allowed is None:
            methods_allowed = ["replace", "add"]

        super(TestValidator, self).__init__(methods_allowed)


class TestJsonPatchMixin(utils.BaseTestCase):
    def test_body_validation(self):
        validator = TestValidator()
        validator.validate_body(
            [{"op": "replace", "path": "/param", "value": "ok"}])
        # invalid if not a list of [{"op": "", "path": "", "value": ""}]
        # is passed
        self.assertRaises(exc.JsonPatchException, validator.validate_body,
                          {"op": "replace", "path": "/me",
                           "value": "should be a list"})

    def test_value_validation(self):
        # a string, a list and a dict are valid value types
        validator = TestValidator()
        validator.validate_body(
            [{"op": "replace", "path": "/param", "value": "ok string"}])
        validator.validate_body(
            [{"op": "replace", "path": "/param",
              "value": ["ok list", "really ok"]}])
        validator.validate_body(
            [{"op": "replace", "path": "/param", "value": {"ok": "dict"}}])

    def test_op_validation(self):
        validator = TestValidator(methods_allowed=["replace", "add", "copy"])
        validator.validate_body(
            [{"op": "copy", "path": "/param", "value": "ok"},
             {"op": "replace", "path": "/param/1", "value": "ok"}])
        self.assertRaises(
            exc.JsonPatchException, validator.validate_body,
            [{"op": "test", "path": "/param", "value": "not allowed"}])
        self.assertRaises(exc.JsonPatchException, validator.validate_body,
                          [{"op": "nosuchmethodatall", "path": "/param",
                           "value": "no way"}])

    def test_path_validation(self):
        validator = TestValidator()
        bad_body_part = {"op": "add", "value": "bad path"}
        for bad_path in ["/param/", "param", "//param", "/param~2", "/param~"]:
            bad_body_part["path"] = bad_path
            bad_body = [bad_body_part]
            self.assertRaises(exc.JsonPatchException,
                              validator.validate_body, bad_body)
        ok_body = [{"op": "add", "value": "some value",
                    "path": "/param~1/param~0"}]
        body = validator.validate_body(ok_body)[0]
        self.assertEqual("param//param~", body["path"])



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

from glance.common import exception as exc
from glance.domain import proxy as image_proxy


def _proxy_artifact_property(attr):
    def getter(self):
        return self.get_type_specific_property(attr)

    def setter(self, value):
        return self.set_type_specific_property(attr, value)

    return property(getter, setter)


class ArtifactHelper(image_proxy.Helper):
    """
    Artifact-friendly proxy helper: does all the same as regular helper
    but also dynamically proxies all the type-specific attributes,
    including properties, blobs and dependencies
    """
    def proxy(self, obj):
        if obj is None or self.proxy_class is None:
            return obj
        if not hasattr(obj, 'metadata'):
            return super(ArtifactHelper, self).proxy(obj)
        extra_attrs = {}
        for att_name in obj.metadata.attributes.all.keys():
            extra_attrs[att_name] = _proxy_artifact_property(att_name)
        new_proxy_class = type("%s(%s)" % (obj.metadata.type_name,
                                           self.proxy_class.__module__),
                               (self.proxy_class,),
                               extra_attrs)
        return new_proxy_class(obj, **self.proxy_kwargs)


class ArtifactRepo(object):
    def __init__(self, base, proxy_helper=None, item_proxy_class=None,
                 item_proxy_kwargs=None):
        self.base = base
        if proxy_helper is None:
            proxy_helper = ArtifactHelper(item_proxy_class, item_proxy_kwargs)
        self.helper = proxy_helper

    def get(self, *args, **kwargs):
        return self.helper.proxy(self.base.get(*args, **kwargs))

    def list(self, *args, **kwargs):
        items = self.base.list(*args, **kwargs)
        return [self.helper.proxy(item) for item in items]

    def add(self, item):
        base_item = self.helper.unproxy(item)
        result = self.base.add(base_item)
        return self.helper.proxy(result)

    def save(self, item):
        base_item = self.helper.unproxy(item)
        result = self.base.save(base_item)
        return self.helper.proxy(result)

    def remove(self, item):
        base_item = self.helper.unproxy(item)
        result = self.base.remove(base_item)
        return self.helper.proxy(result)

    def publish(self, item, *args, **kwargs):
        base_item = self.helper.unproxy(item)
        result = self.base.publish(base_item, *args, **kwargs)
        return self.helper.proxy(result)


class Artifact(object):
    def __init__(self, base, proxy_class=None, proxy_kwargs=None):
        self.base = base
        self.helper = ArtifactHelper(proxy_class, proxy_kwargs)

    # it is enough to proxy metadata only, other properties will be proxied
    # automatically by ArtifactHelper
    metadata = _proxy_artifact_property('metadata')

    def set_type_specific_property(self, prop_name, value):
        setattr(self.base, prop_name, value)

    def get_type_specific_property(self, prop_name):
        try:
            return getattr(self.base, prop_name)
        except AttributeError:
            raise exc.ArtifactInvalidProperty(prop=prop_name)

    def __pre_publish__(self, *args, **kwargs):
        self.base.__pre_publish__(*args, **kwargs)


class ArtifactFactory(object):
    def __init__(self, base,
                 artifact_proxy_class=Artifact,
                 artifact_proxy_kwargs=None):
        self.artifact_helper = ArtifactHelper(artifact_proxy_class,
                                              artifact_proxy_kwargs)
        self.base = base

    def new_artifact(self, *args, **kwargs):
        t = self.base.new_artifact(*args, **kwargs)
        return self.artifact_helper.proxy(t)


class ArtifactBlob(object):
    def __init__(self, base, artifact_blob_proxy_class=None,
                 artifact_blob_proxy_kwargs=None):
        self.base = base
        self.helper = image_proxy.Helper(artifact_blob_proxy_class,
                                         artifact_blob_proxy_kwargs)

    size = _proxy_artifact_property('size')
    locations = _proxy_artifact_property('locations')
    checksum = _proxy_artifact_property('checksum')
    item_key = _proxy_artifact_property('item_key')

    def set_type_specific_property(self, prop_name, value):
        setattr(self.base, prop_name, value)

    def get_type_specific_property(self, prop_name):
        return getattr(self.base, prop_name)

    def to_dict(self):
        return self.base.to_dict()


class ArtifactProperty(object):
    def __init__(self, base, proxy_class=None, proxy_kwargs=None):
        self.base = base
        self.helper = ArtifactHelper(proxy_class, proxy_kwargs)

    def set_type_specific_property(self, prop_name, value):
        setattr(self.base, prop_name, value)

    def get_type_specific_property(self, prop_name):
        return getattr(self.base, prop_name)


class List(collections.MutableSequence):
    def __init__(self, base, item_proxy_class=None,
                 item_proxy_kwargs=None):
        self.base = base
        self.helper = image_proxy.Helper(item_proxy_class, item_proxy_kwargs)

    def __len__(self):
        return len(self.base)

    def __delitem__(self, index):
        del self.base[index]

    def __getitem__(self, index):
        item = self.base[index]
        return self.helper.proxy(item)

    def insert(self, index, value):
        self.base.insert(index, self.helper.unproxy(value))

    def __setitem__(self, index, value):
        self.base[index] = self.helper.unproxy(value)


class Dict(collections.MutableMapping):
    def __init__(self, base, item_proxy_class=None, item_proxy_kwargs=None):
        self.base = base
        self.helper = image_proxy.Helper(item_proxy_class, item_proxy_kwargs)

    def __setitem__(self, key, value):
        self.base[key] = self.helper.unproxy(value)

    def __getitem__(self, key):
        item = self.base[key]
        return self.helper.proxy(item)

    def __delitem__(self, key):
        del self.base[key]

    def __len__(self):
        return len(self.base)

    def __iter__(self):
        for key in self.base.keys():
            yield key



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from glance.common import exception as exc
from glance.glare.domain import proxy
from glance.i18n import _


class ArtifactProxy(proxy.Artifact):
    """A proxy that is capable of modifying an artifact via jsonpatch methods.

    Currently supported methods are update, remove, replace.
    """
    def __init__(self, artifact):
        self.artifact = artifact
        super(ArtifactProxy, self).__init__(artifact)

    def __getattr__(self, name):
        if not hasattr(self, name):
            raise exc.ArtifactInvalidProperty(prop=name)
        return super(ArtifactProxy, self).__getattr__(name)

    def _perform_op(self, op, **kwargs):
        path = kwargs.get("path")
        value = kwargs.get("value")
        prop_name, delimiter, path_left = path.lstrip('/').partition('/')
        super(ArtifactProxy, self).get_type_specific_property(prop_name)
        if not path_left:
            return setattr(self, prop_name, value)
        try:
            prop = self._get_prop_to_update(prop_name, path_left)
            # correct path_left and call corresponding update method
            kwargs["path"] = path_left
            getattr(prop, op)(path=kwargs["path"], value=kwargs.get("value"))
            return setattr(self, prop_name, prop)
        except exc.InvalidJsonPatchPath:
            # NOTE(ivasilevskaya): here exception is reraised with
            # 'part of path' substituted with with 'full path' to form a
            # more relevant message
            raise exc.InvalidJsonPatchPath(
                path=path, explanation=_("No property to access"))

    def _get_prop_to_update(self, prop_name, path):
        """Proxies properties that can be modified via update request.

        All properties can be updated save for 'metadata' and blobs.
        Due to the fact that empty lists and dicts are represented with null
        values, have to check precise type definition by consulting metadata.
        """
        prop = super(ArtifactProxy, self).get_type_specific_property(
            prop_name)
        if (prop_name == "metadata" or
                prop_name in self.artifact.metadata.attributes.blobs):
            return prop
        if not prop:
            # get correct type for empty list/dict
            klass = self.artifact.metadata.attributes.all[prop_name]
            if isinstance(klass, list):
                prop = []
            elif isinstance(klass, dict):
                prop = {}
        return wrap_property(prop, path)

    def replace(self, path, value):
        self._perform_op("replace", path=path, value=value)

    def remove(self, path, value=None):
        self._perform_op("remove", path=path)

    def add(self, path, value):
        self._perform_op("add", path=path, value=value)


class ArtifactFactoryProxy(proxy.ArtifactFactory):
    def __init__(self, factory):
        super(ArtifactFactoryProxy, self).__init__(factory)


class ArtifactRepoProxy(proxy.ArtifactRepo):
    def __init__(self, repo):
        super(ArtifactRepoProxy, self).__init__(
            repo, item_proxy_class=ArtifactProxy)


def wrap_property(prop_value, full_path):
    if isinstance(prop_value, list):
        return ArtifactListPropertyProxy(prop_value, full_path)
    if isinstance(prop_value, dict):
        return ArtifactDictPropertyProxy(prop_value, full_path)
    # no other types are supported
    raise exc.InvalidJsonPatchPath(path=full_path)


class ArtifactListPropertyProxy(proxy.List):
    """A class to wrap a list property.

    Makes possible to modify the property value via supported jsonpatch
    requests (update/remove/replace).
    """
    def __init__(self, prop_value, path):
        super(ArtifactListPropertyProxy, self).__init__(
            prop_value)

    def _proc_key(self, idx_str, should_exist=True):
        """JsonPatchUpdateMixin method overload.

        Only integers less than current array length and '-' (last elem)
        in path are allowed.
        Raises an InvalidJsonPatchPath exception if any of the conditions above
        are not met.
        """
        if idx_str == '-':
            return len(self) - 1
        try:
            idx = int(idx_str)
            if not should_exist and len(self) == 0:
                return 0
            if len(self) < idx + 1:
                msg = _("Array has no element at position %d") % idx
                raise exc.InvalidJsonPatchPath(explanation=msg, path=idx)
            return idx
        except (ValueError, TypeError):
            msg = _("Not an array idx '%s'") % idx_str
            raise exc.InvalidJsonPatchPath(explanation=msg, path=idx_str)

    def add(self, path, value):
        # by now arrays can't contain complex structures (due to Declarative
        # Framework limitations and DB storage model),
        # so will 'path' == idx equality is implied.
        idx = self._proc_key(path, False)
        if idx == len(self) - 1:
            self.append(value)
        else:
            self.insert(idx, value)
        return self.base

    def remove(self, path, value=None):
        # by now arrays can't contain complex structures, so will imply that
        # 'path' == idx [see comment for add()]
        del self[self._proc_key(path)]
        return self.base

    def replace(self, path, value):
        # by now arrays can't contain complex structures, so will imply that
        # 'path' == idx [see comment for add()]
        self[self._proc_key(path)] = value
        return self.base


class ArtifactDictPropertyProxy(proxy.Dict):
    """A class to wrap a dict property.

    Makes possible to modify the property value via supported jsonpatch
    requests (update/remove/replace).
    """
    def __init__(self, prop_value, path):
        super(ArtifactDictPropertyProxy, self).__init__(
            prop_value)

    def _proc_key(self, key_str, should_exist=True):
        """JsonPatchUpdateMixin method overload"""
        if should_exist and key_str not in self.keys():
            msg = _("No such key '%s' in a dict") % key_str
            raise exc.InvalidJsonPatchPath(path=key_str, explanation=msg)
        return key_str

    def replace(self, path, value):
        start, delimiter, rest = path.partition('/')
        # the full path MUST exist in replace operation, so let's check
        # that such key exists
        key = self._proc_key(start)
        if not rest:
            self[key] = value
        else:
            prop = wrap_property(self[key], rest)
            self[key] = prop.replace(rest, value)

    def remove(self, path, value=None):
        start, delimiter, rest = path.partition('/')
        key = self._proc_key(start)
        if not rest:
            del self[key]
        else:
            prop = wrap_property(self[key], rest)
            prop.remove(rest)

    def add(self, path, value):
        start, delimiter, rest = path.partition('/')
        if not rest:
            self[start] = value
        else:
            key = self._proc_key(start)
            prop = wrap_property(self[key], rest)
            self[key] = prop.add(rest, value)



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from glance.common import timeutils
from glance.i18n import _


class Artifact(object):

    def __init__(self, id, name, version, type_name, type_version,
                 state, owner, visibility='private', created_at=None,
                 updated_at=None, **kwargs):
        self.id = id
        self.name = name
        self.type_name = type_name
        self.version = version
        self.type_version = type_version
        self.visibility = visibility
        self.state = state
        self.owner = owner
        self.created_at = created_at
        self.updated_at = updated_at
        self.description = kwargs.pop('description', None)
        self.blobs = kwargs.pop('blobs', {})
        self.properties = kwargs.pop('properties', {})
        self.dependencies = kwargs.pop('dependencies', {})
        self.tags = kwargs.pop('tags', [])

        if kwargs:
            message = _("__init__() got unexpected keyword argument '%s'")
            raise TypeError(message % list(kwargs.keys())[0])


class ArtifactFactory(object):
    def __init__(self, context, klass):
        self.klass = klass
        self.context = context

    def new_artifact(self, name, version, **kwargs):
        id = kwargs.pop('id', str(uuid.uuid4()))
        tags = kwargs.pop('tags', [])
        # pop reserved fields from kwargs dict
        for param in ['owner', 'created_at', 'updated_at',
                      'deleted_at', 'state']:
            kwargs.pop(param, '')
        curr_timestamp = timeutils.utcnow()
        base = self.klass(id=id,
                          name=name,
                          version=version,
                          state='creating',
                          owner=self.context.owner or '',
                          created_at=curr_timestamp,
                          updated_at=curr_timestamp,
                          tags=tags,
                          **kwargs)
        return base



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Sample of plugin for Glance.

For more Glance related benchmarks take a look here:
github.com/openstack/rally/tree/master/samples/tasks/scenarios/glance

About plugins: https://rally.readthedocs.org/en/latest/plugins.html

Rally concepts https://wiki.openstack.org/wiki/Rally/Concepts
"""

import os

from rally.plugins.openstack import scenario
from rally.task import atomic
from rally.task import utils


class GlancePlugin(scenario.OpenStackScenario):

    @atomic.action_timer("glance.create_image_label")
    def _create_image(self, image_name, container_format,
                      image_location, disk_format, **kwargs):
        """Create a new image.

        :param image_name: String used to name the image
        :param container_format: Container format of image.
        Acceptable formats: ami, ari, aki, bare, ovf, ova and docker.
        :param image_location: image file location used to upload
        :param disk_format: Disk format of image. Acceptable formats:
        ami, ari, aki, vhd, vhdx, vmdk, raw, qcow2, vdi, and iso.
        :param **kwargs:  optional parameters to create image

        returns: object of image
        """

        kw = {
            "name": image_name,
            "container_format": container_format,
            "disk_format": disk_format,
        }

        kw.update(kwargs)

        try:
            if os.path.isfile(os.path.expanduser(image_location)):
                kw["data"] = open(os.path.expanduser(image_location))
            else:
                kw["copy_from"] = image_location

            image = self.clients("glance").images.create(**kw)
            image = utils.wait_for(image,
                                   is_ready=utils.resource_is("active"),
                                   update_resource=utils.get_from_manager(),
                                   timeout=100,
                                   check_interval=0.5)
        finally:
            if "data" in kw:
                kw["data"].close()

        return image

    @atomic.action_timer("glance.list_images_label")
    def _list_images(self):
        return list(self.clients("glance").images.list())

    @scenario.configure(context={"cleanup": ["glance"]})
    def create_and_list(self, container_format,
                        image_location, disk_format, **kwargs):
        self._create_image(self.generate_random_name(),
                           container_format,
                           image_location,
                           disk_format,
                           **kwargs)
        self._list_images()



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from six.moves import http_client
import webob.dec

from glance.common import wsgi
from glance.i18n import _, _LW


versions_opts = [

    # Note: Since both glance-api and glare-api have the same name for the
    # option public_endpoint, oslo.config generator throws a DuplicateError
    # exception during the conf file generation incase of differing help
    # texts. Hence we have to have identical help texts for glance-api and
    # glare-api's public_endpoint if not for changing the conf opt name.

    cfg.StrOpt('public_endpoint',
               help=_("""
Public url endpoint to use for Glance/Glare versions response.

This is the public url endpoint that will appear in the Glance/Glare
"versions" response. If no value is specified, the endpoint that is
displayed in the version's response is that of the host running the
API service. Change the endpoint to represent the proxy URL if the
API service is running behind a proxy. If the service is running
behind a load balancer, add the load balancer's URL for this value.

Possible values:
    * None
    * Proxy URL
    * Load balancer URL

Related options:
    * None

""")),
]

CONF = cfg.CONF
CONF.register_opts(versions_opts)

LOG = logging.getLogger(__name__)


class Controller(object):

    """A wsgi controller that reports which API versions are supported."""

    def index(self, req, explicit=False):
        """Respond to a request for all OpenStack API versions."""
        def build_version_object(version, path, status):
            url = CONF.public_endpoint or req.host_url
            return {
                'id': 'v%s' % version,
                'status': status,
                'links': [
                    {
                        'rel': 'self',
                        'href': '%s/%s/' % (url, path),
                    },
                ],
            }

        version_objs = []
        if CONF.enable_v2_api:
            version_objs.extend([
                build_version_object(2.4, 'v2', 'CURRENT'),
                build_version_object(2.3, 'v2', 'SUPPORTED'),
                build_version_object(2.2, 'v2', 'SUPPORTED'),
                build_version_object(2.1, 'v2', 'SUPPORTED'),
                build_version_object(2.0, 'v2', 'SUPPORTED'),
            ])
        if CONF.enable_v1_api:
            LOG.warn(_LW('The Images (Glance) v1 API is deprecated and will '
                         'be removed on or after the Pike release, following '
                         'the standard OpenStack deprecation policy. '
                         'Currently, the solution is to set '
                         'enable_v1_api=False and enable_v2_api=True in your '
                         'glance-api.conf file. Once those options are '
                         'removed from the code, Images (Glance) v2 API will '
                         'be switched on by default and will be the only '
                         'option to deploy and use.'))
            version_objs.extend([
                build_version_object(1.1, 'v1', 'DEPRECATED'),
                build_version_object(1.0, 'v1', 'DEPRECATED'),
            ])

        status = explicit and http_client.OK or http_client.MULTIPLE_CHOICES
        response = webob.Response(request=req,
                                  status=status,
                                  content_type='application/json')
        response.body = jsonutils.dump_as_bytes(dict(versions=version_objs))
        return response

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        return self.index(req)


def create_resource(conf):
    return wsgi.Resource(Controller())



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from glance.api.glare.v0_1 import glare
from glance.common import wsgi


UUID_REGEX = (
    R'[0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}')


class API(wsgi.Router):

    def _get_artifacts_resource(self):
        if not self.artifacts_resource:
            self.artifacts_resource = glare.create_resource()
        return self.artifacts_resource

    def __init__(self, mapper):
        self.artifacts_resource = None
        artifacts_resource = self._get_artifacts_resource()
        reject_method_resource = wsgi.Resource(wsgi.RejectMethodController())

        def _check_json_content_type(environ, result):
            return "application/json" in environ["CONTENT_TYPE"]

        def _check_octet_stream_content_type(environ, result):
            return "application/octet-stream" in environ["CONTENT_TYPE"]

        def connect_routes(m, read_only):
            with m.submapper(resource_name="artifact_operations",
                             path_prefix="/{id}",
                             requirements={'id': UUID_REGEX}) as art:
                art.show()
                if not read_only:
                    art.delete()
                    art.action('update', method='PATCH')
                    art.link('publish', method='POST')

                def connect_attr_action(attr):
                    if not read_only:
                        attr.action("upload", conditions={
                            'method': ["POST", "PUT"],
                            'function': _check_octet_stream_content_type})
                        attr.action("update_property",
                                    conditions={
                                        'method': ["POST", "PUT"],
                                        'function': _check_json_content_type})
                    attr.link("download", method="GET")

                attr_map = art.submapper(resource_name="attr_operations",
                                         path_prefix="/{attr}", path_left=None)
                attr_items = art.submapper(
                    resource_name="attr_item_ops",
                    path_prefix="/{attr}/{path_left:.*}")
                connect_attr_action(attr_map)
                connect_attr_action(attr_items)

            m.connect("", action='list', conditions={'method': 'GET'},
                      state='active')
            m.connect("/drafts", action='list', conditions={'method': 'GET'},
                      state='creating')
            if not read_only:
                m.connect("/drafts", action='create',
                          conditions={'method': 'POST'})

        mapper.connect('/artifacts',
                       controller=artifacts_resource,
                       action='list_artifact_types',
                       conditions={'method': ['GET']})

        versioned = mapper.submapper(path_prefix='/artifacts/{type_name}/'
                                                 'v{type_version}',
                                     controller=artifacts_resource)

        non_versioned = mapper.submapper(path_prefix='/artifacts/{type_name}',
                                         type_version=None,
                                         controller=artifacts_resource)
        connect_routes(versioned, False)
        connect_routes(non_versioned, True)

        mapper.connect('/artifacts',
                       controller=reject_method_resource,
                       action='reject',
                       allowed_methods='GET')

        super(API, self).__init__(mapper)



# Copyright 2010-2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import subprocess

from glance.tests import utils as test_utils


script = """
import os
import sys
# Spoof module installed
sys.modules['%s'] = object
%s
os.environ['EVENTLET_NO_GREENDNS'] = '%s'
if 'eventlet' %s in sys.modules:
    sys.exit(2)
try:
    import glance.cmd
except ImportError:
    sys.exit(%d)
else:
    sys.exit(%d)
"""

eventlet_no_dns = script % ('fake', 'import eventlet', 'foo', 'not', 1, 0)

no_eventlet_no_dns = script % ('fake', '', 'foo', '', 1, 0)

no_eventlet_no_greendns = script % ('dns', '', 'yes', '', 1, 0)

eventlet_no_greendns = script % ('dns', 'import eventlet', 'yes', 'not', 1, 0)

no_eventlet_greendns = script % ('dns', '', 'no', '', 1, 0)

eventlet_greendns = script % ('dns', 'import eventlet', 'no', 'not', 0, 1)


class IPv6ServerTest(test_utils.BaseTestCase):

    def test_no_evnetlet_no_dnspython(self):
        """Test eventlet not imported and dnspython not installed"""
        rc = subprocess.call(['python', '-c', no_eventlet_no_dns])
        self.assertEqual(0, rc)

    def test_evnetlet_no_dnspython(self):
        """Test eventlet pre-imported but dnspython not installed"""
        rc = subprocess.call(['python', '-c', eventlet_no_dns])
        self.assertEqual(0, rc)

    def test_no_eventlet_no_greendns(self):
        """Test eventlet not imported with EVENTLET_NO_GREENDNS='yes'"""
        rc = subprocess.call(['python', '-c', no_eventlet_no_greendns])
        self.assertEqual(0, rc)

    def test_eventlet_no_greendns(self):
        """Test eventlet pre-imported with EVENTLET_NO_GREENDNS='yes'"""
        rc = subprocess.call(['python', '-c', eventlet_no_greendns])
        self.assertEqual(0, rc)

    def test_no_eventlet_w_greendns(self):
        """Test eventlet not imported with EVENTLET_NO_GREENDNS='no'"""
        rc = subprocess.call(['python', '-c', no_eventlet_greendns])
        self.assertEqual(0, rc)

    def test_eventlet_w_greendns(self):
        """Test eventlet pre-imported with EVENTLET_NO_GREENDNS='no'"""
        rc = subprocess.call(['python', '-c', eventlet_greendns])
        self.assertEqual(0, rc)



# Copyright 2010-2011 OpenStack Foundation
# All Rights Reserved.
# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for database migrations run a series of test cases to ensure that
migrations work properly both upgrading and downgrading, and that no data loss
occurs if possible.
"""

from __future__ import print_function

import datetime
import os
import pickle
import uuid

from migrate.versioning import api as migration_api
from migrate.versioning.repository import Repository
from oslo_config import cfg
from oslo_db.sqlalchemy import test_base
from oslo_db.sqlalchemy import test_migrations
from oslo_db.sqlalchemy import utils as db_utils
from oslo_serialization import jsonutils
from oslo_utils import uuidutils
# NOTE(jokke): simplified transition to py3, behaves like py2 xrange
from six.moves import range
import sqlalchemy
import sqlalchemy.types as types

from glance.common import crypt
from glance.common import exception
from glance.common import timeutils
from glance.db import migration
from glance.db.sqlalchemy import migrate_repo
from glance.db.sqlalchemy.migrate_repo.schema import from_migration_import
from glance.db.sqlalchemy.migrate_repo import versions
from glance.db.sqlalchemy import models
from glance.db.sqlalchemy import models_glare
from glance.db.sqlalchemy import models_metadef
import glance.tests.utils as test_utils

from glance.i18n import _

CONF = cfg.CONF
CONF.import_opt('metadata_encryption_key', 'glance.common.config')


def index_exist(index, table, engine):
    inspector = sqlalchemy.inspect(engine)
    return index in [i['name'] for i in inspector.get_indexes(table)]


def unique_constraint_exist(constraint, table, engine):
    inspector = sqlalchemy.inspect(engine)
    return constraint in [c['name'] for c in
                          inspector.get_unique_constraints(table)]


class MigrationsMixin(test_migrations.WalkVersionsMixin):
    @property
    def INIT_VERSION(self):
        return migration.INIT_VERSION

    @property
    def REPOSITORY(self):
        migrate_file = migrate_repo.__file__
        return Repository(os.path.abspath(os.path.dirname(migrate_file)))

    @property
    def migration_api(self):
        return migration_api

    @property
    def migrate_engine(self):
        return self.engine

    def test_walk_versions(self):
        # No more downgrades
        self._walk_versions(False, False)

    def _create_unversioned_001_db(self, engine):
        # Create the initial version of the images table
        meta = sqlalchemy.schema.MetaData()
        meta.bind = engine
        images_001 = sqlalchemy.Table('images', meta,
                                      sqlalchemy.Column('id', models.Integer,
                                                        primary_key=True),
                                      sqlalchemy.Column('name',
                                                        sqlalchemy.String(255)
                                                        ),
                                      sqlalchemy.Column('type',
                                                        sqlalchemy.String(30)),
                                      sqlalchemy.Column('size',
                                                        sqlalchemy.Integer),
                                      sqlalchemy.Column('status',
                                                        sqlalchemy.String(30)),
                                      sqlalchemy.Column('is_public',
                                                        sqlalchemy.Boolean,
                                                        default=False),
                                      sqlalchemy.Column('location',
                                                        sqlalchemy.Text),
                                      sqlalchemy.Column('created_at',
                                                        sqlalchemy.DateTime(),
                                                        nullable=False),
                                      sqlalchemy.Column('updated_at',
                                                        sqlalchemy.DateTime()),
                                      sqlalchemy.Column('deleted_at',
                                                        sqlalchemy.DateTime()),
                                      sqlalchemy.Column('deleted',
                                                        sqlalchemy.Boolean(),
                                                        nullable=False,
                                                        default=False),
                                      mysql_engine='InnoDB',
                                      mysql_charset='utf8')
        images_001.create()

    def test_version_control_existing_db(self):
        """
        Creates a DB without version control information, places it
        under version control and checks that it can be upgraded
        without errors.
        """
        self._create_unversioned_001_db(self.migrate_engine)

        old_version = migration.INIT_VERSION
        # we must start from version 1
        migration.INIT_VERSION = 1
        self.addCleanup(setattr, migration, 'INIT_VERSION', old_version)

        self._walk_versions(False, False)

    def _pre_upgrade_003(self, engine):
        now = datetime.datetime.now()
        images = db_utils.get_table(engine, 'images')
        data = {'deleted': False, 'created_at': now, 'updated_at': now,
                'type': 'kernel', 'status': 'active', 'is_public': True}
        images.insert().values(data).execute()
        return data

    def _check_003(self, engine, data):
        images = db_utils.get_table(engine, 'images')
        self.assertNotIn('type', images.c,
                         "'type' column found in images table columns! "
                         "images table columns reported by metadata: %s\n"
                         % images.c.keys())
        images_prop = db_utils.get_table(engine, 'image_properties')
        result = images_prop.select().execute()
        types = []
        for row in result:
            if row['key'] == 'type':
                types.append(row['value'])
        self.assertIn(data['type'], types)

    def _pre_upgrade_004(self, engine):
        """Insert checksum data sample to check if migration goes fine with
        data.
        """
        now = timeutils.utcnow()
        images = db_utils.get_table(engine, 'images')
        data = [
            {
                'deleted': False, 'created_at': now, 'updated_at': now,
                'type': 'kernel', 'status': 'active', 'is_public': True,
            }
        ]
        engine.execute(images.insert(), data)
        return data

    def _check_004(self, engine, data):
        """Assure that checksum data is present on table"""
        images = db_utils.get_table(engine, 'images')
        self.assertIn('checksum', images.c)
        self.assertEqual(32, images.c['checksum'].type.length)

    def _pre_upgrade_005(self, engine):
        now = timeutils.utcnow()
        images = db_utils.get_table(engine, 'images')
        data = [
            {
                'deleted': False, 'created_at': now, 'updated_at': now,
                'type': 'kernel', 'status': 'active', 'is_public': True,
                # Integer type signed size limit
                'size': 2147483647
            }
        ]
        engine.execute(images.insert(), data)
        return data

    def _check_005(self, engine, data):

        images = db_utils.get_table(engine, 'images')
        select = images.select().execute()

        sizes = [row['size'] for row in select if row['size'] is not None]
        migrated_data_sizes = [element['size'] for element in data]

        for migrated in migrated_data_sizes:
            self.assertIn(migrated, sizes)

    def _pre_upgrade_006(self, engine):
        now = timeutils.utcnow()
        images = db_utils.get_table(engine, 'images')
        image_data = [
            {
                'deleted': False, 'created_at': now, 'updated_at': now,
                'type': 'kernel', 'status': 'active', 'is_public': True,
                'id': 9999,
            }
        ]
        engine.execute(images.insert(), image_data)

        images_properties = db_utils.get_table(engine, 'image_properties')
        properties_data = [
            {
                'id': 10, 'image_id': 9999, 'updated_at': now,
                'created_at': now, 'deleted': False, 'key': 'image_name'
            }
        ]
        engine.execute(images_properties.insert(), properties_data)
        return properties_data

    def _check_006(self, engine, data):
        images_properties = db_utils.get_table(engine, 'image_properties')
        select = images_properties.select().execute()

        # load names from name collumn
        image_names = [row['name'] for row in select]

        # check names from data in image names from name column
        for element in data:
            self.assertIn(element['key'], image_names)

    def _pre_upgrade_010(self, engine):
        """Test rows in images with NULL updated_at get updated to equal
        created_at.
        """

        initial_values = [
            (datetime.datetime(1999, 1, 2, 4, 10, 20),
             datetime.datetime(1999, 1, 2, 4, 10, 30)),
            (datetime.datetime(1999, 2, 4, 6, 15, 25),
             datetime.datetime(1999, 2, 4, 6, 15, 35)),
            (datetime.datetime(1999, 3, 6, 8, 20, 30),
             None),
            (datetime.datetime(1999, 4, 8, 10, 25, 35),
             None),
        ]

        images = db_utils.get_table(engine, 'images')
        for created_at, updated_at in initial_values:
            row = dict(deleted=False,
                       created_at=created_at,
                       updated_at=updated_at,
                       status='active',
                       is_public=True,
                       min_disk=0,
                       min_ram=0)
            images.insert().values(row).execute()

        return initial_values

    def _check_010(self, engine, data):
        values = {c: u for c, u in data}

        images = db_utils.get_table(engine, 'images')
        for row in images.select().execute():
            if row['created_at'] in values:
                # updated_at should be unchanged if not previous NULL, or
                # set to created_at if previously NULL
                updated_at = values.pop(row['created_at']) or row['created_at']
                self.assertEqual(row['updated_at'], updated_at)

        # No initial values should be remaining
        self.assertEqual(0, len(values))

    def _pre_upgrade_012(self, engine):
        """Test rows in images have id changes from int to varchar(32) and
        value changed from int to UUID. Also test image_members and
        image_properties gets updated to point to new UUID keys.
        """

        images = db_utils.get_table(engine, 'images')
        image_members = db_utils.get_table(engine, 'image_members')
        image_properties = db_utils.get_table(engine, 'image_properties')

        # Insert kernel, ramdisk and normal images
        now = timeutils.utcnow()
        data = {'created_at': now, 'updated_at': now,
                'status': 'active', 'deleted': False,
                'is_public': True, 'min_disk': 0, 'min_ram': 0}

        test_data = {}
        for name in ('kernel', 'ramdisk', 'normal'):
            data['name'] = '%s migration 012 test' % name
            result = images.insert().values(data).execute()
            test_data[name] = result.inserted_primary_key[0]

        # Insert image_members and image_properties rows
        data = {'created_at': now, 'updated_at': now, 'deleted': False,
                'image_id': test_data['normal'], 'member': 'foobar',
                'can_share': False}
        result = image_members.insert().values(data).execute()
        test_data['member'] = result.inserted_primary_key[0]

        data = {'created_at': now, 'updated_at': now, 'deleted': False,
                'image_id': test_data['normal'], 'name': 'ramdisk_id',
                'value': test_data['ramdisk']}
        result = image_properties.insert().values(data).execute()
        test_data['properties'] = [result.inserted_primary_key[0]]

        data.update({'name': 'kernel_id', 'value': test_data['kernel']})
        result = image_properties.insert().values(data).execute()
        test_data['properties'].append(result.inserted_primary_key)

        return test_data

    def _check_012(self, engine, test_data):
        images = db_utils.get_table(engine, 'images')
        image_members = db_utils.get_table(engine, 'image_members')
        image_properties = db_utils.get_table(engine, 'image_properties')

        # Find kernel, ramdisk and normal images. Make sure id has been
        # changed to a uuid
        uuids = {}
        for name in ('kernel', 'ramdisk', 'normal'):
            image_name = '%s migration 012 test' % name
            rows = images.select().where(
                images.c.name == image_name).execute().fetchall()

            self.assertEqual(1, len(rows))

            row = rows[0]
            self.assertTrue(uuidutils.is_uuid_like(row['id']))

            uuids[name] = row['id']

        # Find all image_members to ensure image_id has been updated
        results = image_members.select().where(
            image_members.c.image_id == uuids['normal']).execute().fetchall()
        self.assertEqual(1, len(results))

        # Find all image_properties to ensure image_id has been updated
        # as well as ensure kernel_id and ramdisk_id values have been
        # updated too
        results = image_properties.select().where(
            image_properties.c.image_id == uuids['normal']
        ).execute().fetchall()
        self.assertEqual(2, len(results))
        for row in results:
            self.assertIn(row['name'], ('kernel_id', 'ramdisk_id'))

            if row['name'] == 'kernel_id':
                self.assertEqual(row['value'], uuids['kernel'])
            if row['name'] == 'ramdisk_id':
                self.assertEqual(row['value'], uuids['ramdisk'])

    def _assert_invalid_swift_uri_raises_bad_store_uri(self,
                                                       legacy_parse_uri_fn):
        invalid_uri = ('swift://http://acct:usr:pass@example.com'
                       '/container/obj-id')
        # URI cannot contain more than one occurrence of a scheme.
        self.assertRaises(exception.BadStoreUri,
                          legacy_parse_uri_fn,
                          invalid_uri,
                          True)

        invalid_scheme_uri = ('http://acct:usr:pass@example.com'
                              '/container/obj-id')
        self.assertRaises(exception.BadStoreUri,
                          legacy_parse_uri_fn,
                          invalid_scheme_uri,
                          True)

        invalid_account_missing_uri = 'swift+http://container/obj-id'
        # Badly formed Swift URI: swift+http://container/obj-id
        self.assertRaises(exception.BadStoreUri,
                          legacy_parse_uri_fn,
                          invalid_account_missing_uri,
                          True)

        invalid_container_missing_uri = ('swift+http://'
                                         'acct:usr:pass@example.com/obj-id')
        # Badly formed Swift URI: swift+http://acct:usr:pass@example.com/obj-id
        self.assertRaises(exception.BadStoreUri,
                          legacy_parse_uri_fn,
                          invalid_container_missing_uri,
                          True)

        invalid_object_missing_uri = ('swift+http://'
                                      'acct:usr:pass@example.com/container')
        # Badly formed Swift URI:
        # swift+http://acct:usr:pass@example.com/container
        self.assertRaises(exception.BadStoreUri,
                          legacy_parse_uri_fn,
                          invalid_object_missing_uri,
                          True)

        invalid_user_without_pass_uri = ('swift://acctusr@example.com'
                                         '/container/obj-id')
        # Badly formed credentials '%(creds)s' in Swift URI
        self.assertRaises(exception.BadStoreUri,
                          legacy_parse_uri_fn,
                          invalid_user_without_pass_uri,
                          True)

        # Badly formed credentials in Swift URI.
        self.assertRaises(exception.BadStoreUri,
                          legacy_parse_uri_fn,
                          invalid_user_without_pass_uri,
                          False)

    def test_legacy_parse_swift_uri_015(self):
        (legacy_parse_uri,) = from_migration_import(
            '015_quote_swift_credentials', ['legacy_parse_uri'])

        uri = legacy_parse_uri(
            'swift://acct:usr:pass@example.com/container/obj-id',
            True)
        self.assertTrue(uri, 'swift://acct%3Ausr:pass@example.com'
                             '/container/obj-id')

        self._assert_invalid_swift_uri_raises_bad_store_uri(legacy_parse_uri)

    def _pre_upgrade_015(self, engine):
        images = db_utils.get_table(engine, 'images')
        unquoted_locations = [
            'swift://acct:usr:pass@example.com/container/obj-id',
            'file://foo',
        ]
        now = datetime.datetime.now()
        temp = dict(deleted=False,
                    created_at=now,
                    updated_at=now,
                    status='active',
                    is_public=True,
                    min_disk=0,
                    min_ram=0)
        data = []
        for i, location in enumerate(unquoted_locations):
            temp.update(location=location, id=str(uuid.uuid4()))
            data.append(temp)
            images.insert().values(temp).execute()
        return data

    def _check_015(self, engine, data):
        images = db_utils.get_table(engine, 'images')
        quoted_locations = [
            'swift://acct%3Ausr:pass@example.com/container/obj-id',
            'file://foo',
        ]
        result = images.select().execute()
        locations = list(map(lambda x: x['location'], result))
        for loc in quoted_locations:
            self.assertIn(loc, locations)

    def _pre_upgrade_016(self, engine):
        images = db_utils.get_table(engine, 'images')
        now = datetime.datetime.now()
        temp = dict(deleted=False,
                    created_at=now,
                    updated_at=now,
                    status='active',
                    is_public=True,
                    min_disk=0,
                    min_ram=0,
                    id='fake-image-id1')
        images.insert().values(temp).execute()
        image_members = db_utils.get_table(engine, 'image_members')
        now = datetime.datetime.now()
        data = {'deleted': False,
                'created_at': now,
                'member': 'fake-member',
                'updated_at': now,
                'can_share': False,
                'image_id': 'fake-image-id1'}
        image_members.insert().values(data).execute()
        return data

    def _check_016(self, engine, data):
        image_members = db_utils.get_table(engine, 'image_members')
        self.assertIn('status', image_members.c,
                      "'status' column found in image_members table "
                      "columns! image_members table columns: %s"
                      % image_members.c.keys())

    def test_legacy_parse_swift_uri_017(self):
        metadata_encryption_key = 'a' * 16
        CONF.set_override('metadata_encryption_key', metadata_encryption_key,
                          enforce_type=True)
        self.addCleanup(CONF.reset)
        (legacy_parse_uri, encrypt_location) = from_migration_import(
            '017_quote_encrypted_swift_credentials', ['legacy_parse_uri',
                                                      'encrypt_location'])

        uri = legacy_parse_uri('swift://acct:usr:pass@example.com'
                               '/container/obj-id', True)
        self.assertTrue(uri, encrypt_location(
            'swift://acct%3Ausr:pass@example.com/container/obj-id'))

        self._assert_invalid_swift_uri_raises_bad_store_uri(legacy_parse_uri)

    def _pre_upgrade_017(self, engine):
        metadata_encryption_key = 'a' * 16
        CONF.set_override('metadata_encryption_key', metadata_encryption_key,
                          enforce_type=True)
        self.addCleanup(CONF.reset)
        images = db_utils.get_table(engine, 'images')
        unquoted = 'swift://acct:usr:pass@example.com/container/obj-id'
        encrypted_unquoted = crypt.urlsafe_encrypt(
            metadata_encryption_key,
            unquoted, 64)
        data = []
        now = datetime.datetime.now()
        temp = dict(deleted=False,
                    created_at=now,
                    updated_at=now,
                    status='active',
                    is_public=True,
                    min_disk=0,
                    min_ram=0,
                    location=encrypted_unquoted,
                    id='fakeid1')
        images.insert().values(temp).execute()

        locations = [
            'file://ab',
            'file://abc',
            'swift://acct3A%foobar:pass@example.com/container/obj-id2'
        ]

        now = datetime.datetime.now()
        temp = dict(deleted=False,
                    created_at=now,
                    updated_at=now,
                    status='active',
                    is_public=True,
                    min_disk=0,
                    min_ram=0)
        for i, location in enumerate(locations):
            temp.update(location=location, id=str(uuid.uuid4()))
            data.append(temp)
            images.insert().values(temp).execute()
        return data

    def _check_017(self, engine, data):
        metadata_encryption_key = 'a' * 16
        quoted = 'swift://acct%3Ausr:pass@example.com/container/obj-id'
        images = db_utils.get_table(engine, 'images')
        result = images.select().execute()
        locations = list(map(lambda x: x['location'], result))
        actual_location = []
        for location in locations:
            if location:
                try:
                    temp_loc = crypt.urlsafe_decrypt(metadata_encryption_key,
                                                     location)
                    actual_location.append(temp_loc)
                except TypeError:
                    actual_location.append(location)
                except ValueError:
                    actual_location.append(location)

        self.assertIn(quoted, actual_location)
        loc_list = ['file://ab',
                    'file://abc',
                    'swift://acct3A%foobar:pass@example.com/container/obj-id2']

        for location in loc_list:
            if location not in actual_location:
                self.fail(_("location: %s data lost") % location)

    def _pre_upgrade_019(self, engine):
        images = db_utils.get_table(engine, 'images')
        now = datetime.datetime.now()
        base_values = {
            'deleted': False,
            'created_at': now,
            'updated_at': now,
            'status': 'active',
            'is_public': True,
            'min_disk': 0,
            'min_ram': 0,
        }
        data = [
            {'id': 'fake-19-1', 'location': 'http://glance.example.com'},
            # NOTE(bcwaldon): images with a location of None should
            # not be migrated
            {'id': 'fake-19-2', 'location': None},
        ]
        for image in data:
            image.update(base_values)
            images.insert().values(image).execute()
        return data

    def _check_019(self, engine, data):
        image_locations = db_utils.get_table(engine, 'image_locations')
        records = image_locations.select().execute().fetchall()
        locations = {il.image_id: il.value for il in records}
        self.assertEqual('http://glance.example.com',
                         locations.get('fake-19-1'))

    def _check_020(self, engine, data):
        images = db_utils.get_table(engine, 'images')
        self.assertNotIn('location', images.c)

    def _pre_upgrade_026(self, engine):
        image_locations = db_utils.get_table(engine, 'image_locations')

        now = datetime.datetime.now()
        image_id = 'fake_id'
        url = 'file:///some/place/onthe/fs'

        images = db_utils.get_table(engine, 'images')
        temp = dict(deleted=False,
                    created_at=now,
                    updated_at=now,
                    status='active',
                    is_public=True,
                    min_disk=0,
                    min_ram=0,
                    id=image_id)
        images.insert().values(temp).execute()

        temp = dict(deleted=False,
                    created_at=now,
                    updated_at=now,
                    image_id=image_id,
                    value=url)
        image_locations.insert().values(temp).execute()
        return image_id

    def _check_026(self, engine, data):
        image_locations = db_utils.get_table(engine, 'image_locations')
        results = image_locations.select().where(
            image_locations.c.image_id == data).execute()

        r = list(results)
        self.assertEqual(1, len(r))
        self.assertEqual('file:///some/place/onthe/fs', r[0]['value'])
        self.assertIn('meta_data', r[0])
        x = pickle.loads(r[0]['meta_data'])
        self.assertEqual({}, x)

    def _check_027(self, engine, data):
        table = "images"
        index = "checksum_image_idx"
        columns = ["checksum"]

        meta = sqlalchemy.MetaData()
        meta.bind = engine

        new_table = sqlalchemy.Table(table, meta, autoload=True)

        index_data = [(idx.name, idx.columns.keys())
                      for idx in new_table.indexes]

        self.assertIn((index, columns), index_data)

    def _check_028(self, engine, data):
        owner_index = "owner_image_idx"
        columns = ["owner"]

        images_table = db_utils.get_table(engine, 'images')

        index_data = [(idx.name, idx.columns.keys())
                      for idx in images_table.indexes
                      if idx.name == owner_index]

        self.assertIn((owner_index, columns), index_data)

    def _pre_upgrade_029(self, engine):
        image_locations = db_utils.get_table(engine, 'image_locations')

        meta_data = {'somelist': ['a', 'b', 'c'], 'avalue': 'hello',
                     'adict': {}}

        now = datetime.datetime.now()
        image_id = 'fake_029_id'
        url = 'file:///some/place/onthe/fs029'

        images = db_utils.get_table(engine, 'images')
        temp = dict(deleted=False,
                    created_at=now,
                    updated_at=now,
                    status='active',
                    is_public=True,
                    min_disk=0,
                    min_ram=0,
                    id=image_id)
        images.insert().values(temp).execute()

        pickle_md = pickle.dumps(meta_data)
        temp = dict(deleted=False,
                    created_at=now,
                    updated_at=now,
                    image_id=image_id,
                    value=url,
                    meta_data=pickle_md)
        image_locations.insert().values(temp).execute()

        return meta_data, image_id

    def _check_029(self, engine, data):
        meta_data = data[0]
        image_id = data[1]
        image_locations = db_utils.get_table(engine, 'image_locations')

        records = image_locations.select().where(
            image_locations.c.image_id == image_id).execute().fetchall()

        for r in records:
            d = jsonutils.loads(r['meta_data'])
            self.assertEqual(d, meta_data)

    def _check_030(self, engine, data):
        table = "tasks"
        index_type = ('ix_tasks_type', ['type'])
        index_status = ('ix_tasks_status', ['status'])
        index_owner = ('ix_tasks_owner', ['owner'])
        index_deleted = ('ix_tasks_deleted', ['deleted'])
        index_updated_at = ('ix_tasks_updated_at', ['updated_at'])

        meta = sqlalchemy.MetaData()
        meta.bind = engine

        tasks_table = sqlalchemy.Table(table, meta, autoload=True)

        index_data = [(idx.name, idx.columns.keys())
                      for idx in tasks_table.indexes]

        self.assertIn(index_type, index_data)
        self.assertIn(index_status, index_data)
        self.assertIn(index_owner, index_data)
        self.assertIn(index_deleted, index_data)
        self.assertIn(index_updated_at, index_data)

        expected = [u'id',
                    u'type',
                    u'status',
                    u'owner',
                    u'input',
                    u'result',
                    u'message',
                    u'expires_at',
                    u'created_at',
                    u'updated_at',
                    u'deleted_at',
                    u'deleted']

        # NOTE(flwang): Skip the column type checking for now since Jenkins is
        # using sqlalchemy.dialects.postgresql.base.TIMESTAMP instead of
        # DATETIME which is using by mysql and sqlite.
        col_data = [col.name for col in tasks_table.columns]
        self.assertEqual(expected, col_data)

    def _pre_upgrade_031(self, engine):
        images = db_utils.get_table(engine, 'images')
        now = datetime.datetime.now()
        image_id = 'fake_031_id'
        temp = dict(deleted=False,
                    created_at=now,
                    updated_at=now,
                    status='active',
                    is_public=True,
                    min_disk=0,
                    min_ram=0,
                    id=image_id)
        images.insert().values(temp).execute()

        locations_table = db_utils.get_table(engine, 'image_locations')
        locations = [
            ('file://ab', '{"a": "yo yo"}'),
            ('file://ab', '{}'),
            ('file://ab', '{}'),
            ('file://ab1', '{"a": "that one, please"}'),
            ('file://ab1', '{"a": "that one, please"}'),
        ]
        temp = dict(deleted=False,
                    created_at=now,
                    updated_at=now,
                    image_id=image_id)

        for location, metadata in locations:
            temp.update(value=location, meta_data=metadata)
            locations_table.insert().values(temp).execute()
        return image_id

    def _check_031(self, engine, image_id):
        locations_table = db_utils.get_table(engine, 'image_locations')
        result = locations_table.select().where(
            locations_table.c.image_id == image_id).execute().fetchall()

        locations = set([(x['value'], x['meta_data']) for x in result])
        actual_locations = set([
            ('file://ab', '{"a": "yo yo"}'),
            ('file://ab', '{}'),
            ('file://ab1', '{"a": "that one, please"}'),
        ])
        self.assertFalse(actual_locations.symmetric_difference(locations))

    def _pre_upgrade_032(self, engine):
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine, 'task_info')

        tasks = db_utils.get_table(engine, 'tasks')
        now = datetime.datetime.now()
        base_values = {
            'deleted': False,
            'created_at': now,
            'updated_at': now,
            'status': 'active',
            'owner': 'TENANT',
            'type': 'import',
        }
        data = [
            {
                'id': 'task-1',
                'input': 'some input',
                'message': None,
                'result': 'successful'
            },
            {
                'id': 'task-2',
                'input': None,
                'message': None,
                'result': None
            },
        ]
        for task in data:
            task.update(base_values)
            tasks.insert().values(task).execute()
        return data

    def _check_032(self, engine, data):
        task_info_table = db_utils.get_table(engine, 'task_info')

        task_info_refs = task_info_table.select().execute().fetchall()

        self.assertEqual(2, len(task_info_refs))

        for x in range(len(task_info_refs)):
            self.assertEqual(task_info_refs[x].task_id, data[x]['id'])
            self.assertEqual(task_info_refs[x].input, data[x]['input'])
            self.assertEqual(task_info_refs[x].result, data[x]['result'])
            self.assertIsNone(task_info_refs[x].message)

        tasks_table = db_utils.get_table(engine, 'tasks')
        self.assertNotIn('input', tasks_table.c)
        self.assertNotIn('result', tasks_table.c)
        self.assertNotIn('message', tasks_table.c)

    def _pre_upgrade_033(self, engine):
        images = db_utils.get_table(engine, 'images')
        image_locations = db_utils.get_table(engine, 'image_locations')

        now = datetime.datetime.now()
        image_id = 'fake_id_028_%d'
        url = 'file:///some/place/onthe/fs_%d'
        status_list = ['active', 'saving', 'queued', 'killed',
                       'pending_delete', 'deleted']
        image_id_list = []

        for (idx, status) in enumerate(status_list):
            temp = dict(deleted=False,
                        created_at=now,
                        updated_at=now,
                        status=status,
                        is_public=True,
                        min_disk=0,
                        min_ram=0,
                        id=image_id % idx)
            images.insert().values(temp).execute()

            temp = dict(deleted=False,
                        created_at=now,
                        updated_at=now,
                        image_id=image_id % idx,
                        value=url % idx)
            image_locations.insert().values(temp).execute()

            image_id_list.append(image_id % idx)
        return image_id_list

    def _check_033(self, engine, data):
        image_locations = db_utils.get_table(engine, 'image_locations')

        self.assertIn('status', image_locations.c)
        self.assertEqual(30, image_locations.c['status'].type.length)

        status_list = ['active', 'active', 'active',
                       'deleted', 'pending_delete', 'deleted']

        for (idx, image_id) in enumerate(data):
            results = image_locations.select().where(
                image_locations.c.image_id == image_id).execute()
            r = list(results)
            self.assertEqual(1, len(r))
            self.assertIn('status', r[0])
            self.assertEqual(status_list[idx], r[0]['status'])

    def _pre_upgrade_034(self, engine):
        images = db_utils.get_table(engine, 'images')

        now = datetime.datetime.now()
        image_id = 'fake_id_034'
        temp = dict(deleted=False,
                    created_at=now,
                    status='active',
                    is_public=True,
                    min_disk=0,
                    min_ram=0,
                    id=image_id)
        images.insert().values(temp).execute()

    def _check_034(self, engine, data):
        images = db_utils.get_table(engine, 'images')
        self.assertIn('virtual_size', images.c)

        result = (images.select()
                  .where(images.c.id == 'fake_id_034')
                  .execute().fetchone())
        self.assertIsNone(result.virtual_size)

    def _pre_upgrade_035(self, engine):
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine, 'metadef_namespaces')
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine, 'metadef_properties')
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine, 'metadef_objects')
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine, 'metadef_resource_types')
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine,
                          'metadef_namespace_resource_types')

    def _check_035(self, engine, data):
        meta = sqlalchemy.MetaData()
        meta.bind = engine

        # metadef_namespaces
        table = sqlalchemy.Table("metadef_namespaces", meta, autoload=True)
        index_namespace = ('ix_namespaces_namespace', ['namespace'])
        index_data = [(idx.name, idx.columns.keys())
                      for idx in table.indexes]
        self.assertIn(index_namespace, index_data)

        expected_cols = [u'id',
                         u'namespace',
                         u'display_name',
                         u'description',
                         u'visibility',
                         u'protected',
                         u'owner',
                         u'created_at',
                         u'updated_at']
        col_data = [col.name for col in table.columns]
        self.assertEqual(expected_cols, col_data)

        # metadef_objects
        table = sqlalchemy.Table("metadef_objects", meta, autoload=True)
        index_namespace_id_name = (
            'ix_objects_namespace_id_name', ['namespace_id', 'name'])
        index_data = [(idx.name, idx.columns.keys())
                      for idx in table.indexes]
        self.assertIn(index_namespace_id_name, index_data)

        expected_cols = [u'id',
                         u'namespace_id',
                         u'name',
                         u'description',
                         u'required',
                         u'schema',
                         u'created_at',
                         u'updated_at']
        col_data = [col.name for col in table.columns]
        self.assertEqual(expected_cols, col_data)

        # metadef_properties
        table = sqlalchemy.Table("metadef_properties", meta, autoload=True)
        index_namespace_id_name = (
            'ix_metadef_properties_namespace_id_name',
            ['namespace_id', 'name'])
        index_data = [(idx.name, idx.columns.keys())
                      for idx in table.indexes]
        self.assertIn(index_namespace_id_name, index_data)

        expected_cols = [u'id',
                         u'namespace_id',
                         u'name',
                         u'schema',
                         u'created_at',
                         u'updated_at']
        col_data = [col.name for col in table.columns]
        self.assertEqual(expected_cols, col_data)

        # metadef_resource_types
        table = sqlalchemy.Table(
            "metadef_resource_types", meta, autoload=True)
        index_resource_types_name = (
            'ix_metadef_resource_types_name', ['name'])
        index_data = [(idx.name, idx.columns.keys())
                      for idx in table.indexes]
        self.assertIn(index_resource_types_name, index_data)

        expected_cols = [u'id',
                         u'name',
                         u'protected',
                         u'created_at',
                         u'updated_at']
        col_data = [col.name for col in table.columns]
        self.assertEqual(expected_cols, col_data)

        # metadef_namespace_resource_types
        table = sqlalchemy.Table(
            "metadef_namespace_resource_types", meta, autoload=True)
        index_ns_res_types_res_type_id_ns_id = (
            'ix_metadef_ns_res_types_res_type_id_ns_id',
            ['resource_type_id', 'namespace_id'])
        index_data = [(idx.name, idx.columns.keys())
                      for idx in table.indexes]
        self.assertIn(index_ns_res_types_res_type_id_ns_id, index_data)

        expected_cols = [u'resource_type_id',
                         u'namespace_id',
                         u'properties_target',
                         u'prefix',
                         u'created_at',
                         u'updated_at']
        col_data = [col.name for col in table.columns]
        self.assertEqual(expected_cols, col_data)

    def _pre_upgrade_036(self, engine):
        meta = sqlalchemy.MetaData()
        meta.bind = engine

        # metadef_objects
        table = sqlalchemy.Table("metadef_objects", meta, autoload=True)
        expected_cols = [u'id',
                         u'namespace_id',
                         u'name',
                         u'description',
                         u'required',
                         u'schema',
                         u'created_at',
                         u'updated_at']
        col_data = [col.name for col in table.columns]
        self.assertEqual(expected_cols, col_data)

        # metadef_properties
        table = sqlalchemy.Table("metadef_properties", meta, autoload=True)
        expected_cols = [u'id',
                         u'namespace_id',
                         u'name',
                         u'schema',
                         u'created_at',
                         u'updated_at']
        col_data = [col.name for col in table.columns]
        self.assertEqual(expected_cols, col_data)

    def _check_036(self, engine, data):
        meta = sqlalchemy.MetaData()
        meta.bind = engine

        # metadef_objects
        table = sqlalchemy.Table("metadef_objects", meta, autoload=True)
        expected_cols = [u'id',
                         u'namespace_id',
                         u'name',
                         u'description',
                         u'required',
                         u'json_schema',
                         u'created_at',
                         u'updated_at']
        col_data = [col.name for col in table.columns]
        self.assertEqual(expected_cols, col_data)

        # metadef_properties
        table = sqlalchemy.Table("metadef_properties", meta, autoload=True)
        expected_cols = [u'id',
                         u'namespace_id',
                         u'name',
                         u'json_schema',
                         u'created_at',
                         u'updated_at']
        col_data = [col.name for col in table.columns]
        self.assertEqual(expected_cols, col_data)

    def _check_037(self, engine, data):
        if engine.name == 'mysql':
            self.assertFalse(unique_constraint_exist('image_id',
                                                     'image_properties',
                                                     engine))

            self.assertTrue(unique_constraint_exist(
                'ix_image_properties_image_id_name',
                'image_properties',
                engine))

        image_members = db_utils.get_table(engine, 'image_members')
        images = db_utils.get_table(engine, 'images')

        self.assertFalse(image_members.c.status.nullable)
        self.assertFalse(images.c.protected.nullable)

        now = datetime.datetime.now()
        temp = dict(
            deleted=False,
            created_at=now,
            status='active',
            is_public=True,
            min_disk=0,
            min_ram=0,
            id='fake_image_035'
        )
        images.insert().values(temp).execute()

        image = (images.select()
                 .where(images.c.id == 'fake_image_035')
                 .execute().fetchone())

        self.assertFalse(image['protected'])

        temp = dict(
            deleted=False,
            created_at=now,
            image_id='fake_image_035',
            member='fake_member',
            can_share=True,
            id=3
        )

        image_members.insert().values(temp).execute()

        image_member = (image_members.select()
                        .where(image_members.c.id == 3)
                        .execute().fetchone())

        self.assertEqual('pending', image_member['status'])

    def _pre_upgrade_038(self, engine):
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine, 'metadef_tags')

    def _check_038(self, engine, data):
        meta = sqlalchemy.MetaData()
        meta.bind = engine

        # metadef_tags
        table = sqlalchemy.Table("metadef_tags", meta, autoload=True)
        expected_cols = [u'id',
                         u'namespace_id',
                         u'name',
                         u'created_at',
                         u'updated_at']
        col_data = [col.name for col in table.columns]
        self.assertEqual(expected_cols, col_data)

    def _check_039(self, engine, data):
        meta = sqlalchemy.MetaData()
        meta.bind = engine

        metadef_namespaces = sqlalchemy.Table('metadef_namespaces', meta,
                                              autoload=True)
        metadef_properties = sqlalchemy.Table('metadef_properties', meta,
                                              autoload=True)
        metadef_objects = sqlalchemy.Table('metadef_objects', meta,
                                           autoload=True)
        metadef_ns_res_types = sqlalchemy.Table(
            'metadef_namespace_resource_types',
            meta, autoload=True)
        metadef_resource_types = sqlalchemy.Table('metadef_resource_types',
                                                  meta, autoload=True)

        tables = [metadef_namespaces, metadef_properties, metadef_objects,
                  metadef_ns_res_types, metadef_resource_types]

        for table in tables:
            for index_name in ['ix_namespaces_namespace',
                               'ix_objects_namespace_id_name',
                               'ix_metadef_properties_namespace_id_name']:
                self.assertFalse(index_exist(index_name, table.name, engine))
            for uc_name in ['resource_type_id', 'namespace', 'name',
                            'namespace_id',
                            'metadef_objects_namespace_id_name_key',
                            'metadef_properties_namespace_id_name_key']:
                self.assertFalse(unique_constraint_exist(uc_name, table.name,
                                                         engine))

        self.assertTrue(index_exist('ix_metadef_ns_res_types_namespace_id',
                                    metadef_ns_res_types.name, engine))

        self.assertTrue(index_exist('ix_metadef_namespaces_namespace',
                                    metadef_namespaces.name, engine))

        self.assertTrue(index_exist('ix_metadef_namespaces_owner',
                                    metadef_namespaces.name, engine))

        self.assertTrue(index_exist('ix_metadef_objects_name',
                                    metadef_objects.name, engine))

        self.assertTrue(index_exist('ix_metadef_objects_namespace_id',
                                    metadef_objects.name, engine))

        self.assertTrue(index_exist('ix_metadef_properties_name',
                                    metadef_properties.name, engine))

        self.assertTrue(index_exist('ix_metadef_properties_namespace_id',
                                    metadef_properties.name, engine))

    def _check_040(self, engine, data):
        meta = sqlalchemy.MetaData()
        meta.bind = engine
        metadef_tags = sqlalchemy.Table('metadef_tags', meta, autoload=True)

        if engine.name == 'mysql':
            self.assertFalse(index_exist('namespace_id',
                             metadef_tags.name, engine))

    def _pre_upgrade_041(self, engine):
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine,
                          'artifacts')
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine,
                          'artifact_tags')
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine,
                          'artifact_properties')
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine,
                          'artifact_blobs')
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine,
                          'artifact_dependencies')
        self.assertRaises(sqlalchemy.exc.NoSuchTableError,
                          db_utils.get_table, engine,
                          'artifact_locations')

    def _check_041(self, engine, data):
        artifacts_indices = [('ix_artifact_name_and_version',
                              ['name', 'version_prefix', 'version_suffix']),
                             ('ix_artifact_type',
                              ['type_name',
                               'type_version_prefix',
                               'type_version_suffix']),
                             ('ix_artifact_state', ['state']),
                             ('ix_artifact_visibility', ['visibility']),
                             ('ix_artifact_owner', ['owner'])]
        artifacts_columns = ['id',
                             'name',
                             'type_name',
                             'type_version_prefix',
                             'type_version_suffix',
                             'type_version_meta',
                             'version_prefix',
                             'version_suffix',
                             'version_meta',
                             'description',
                             'visibility',
                             'state',
                             'owner',
                             'created_at',
                             'updated_at',
                             'deleted_at',
                             'published_at']
        self.assert_table(engine, 'artifacts', artifacts_indices,
                          artifacts_columns)

        tags_indices = [('ix_artifact_tags_artifact_id', ['artifact_id']),
                        ('ix_artifact_tags_artifact_id_tag_value',
                         ['artifact_id',
                          'value'])]
        tags_columns = ['id',
                        'artifact_id',
                        'value',
                        'created_at',
                        'updated_at']
        self.assert_table(engine, 'artifact_tags', tags_indices, tags_columns)

        prop_indices = [
            ('ix_artifact_properties_artifact_id', ['artifact_id']),
            ('ix_artifact_properties_name', ['name'])]
        prop_columns = ['id',
                        'artifact_id',
                        'name',
                        'string_value',
                        'int_value',
                        'numeric_value',
                        'bool_value',
                        'text_value',
                        'created_at',
                        'updated_at',
                        'position']
        self.assert_table(engine, 'artifact_properties', prop_indices,
                          prop_columns)

        blobs_indices = [
            ('ix_artifact_blobs_artifact_id', ['artifact_id']),
            ('ix_artifact_blobs_name', ['name'])]
        blobs_columns = ['id',
                         'artifact_id',
                         'size',
                         'checksum',
                         'name',
                         'item_key',
                         'position',
                         'created_at',
                         'updated_at']
        self.assert_table(engine, 'artifact_blobs', blobs_indices,
                          blobs_columns)

        dependencies_indices = [
            ('ix_artifact_dependencies_source_id', ['artifact_source']),
            ('ix_artifact_dependencies_direct_dependencies',
             ['artifact_source', 'is_direct']),
            ('ix_artifact_dependencies_dest_id', ['artifact_dest']),
            ('ix_artifact_dependencies_origin_id', ['artifact_origin'])]
        dependencies_columns = ['id',
                                'artifact_source',
                                'artifact_dest',
                                'artifact_origin',
                                'is_direct',
                                'position',
                                'name',
                                'created_at',
                                'updated_at']
        self.assert_table(engine, 'artifact_dependencies',
                          dependencies_indices,
                          dependencies_columns)

        locations_indices = [
            ('ix_artifact_blob_locations_blob_id', ['blob_id'])]
        locations_columns = ['id',
                             'blob_id',
                             'value',
                             'created_at',
                             'updated_at',
                             'position',
                             'status']
        self.assert_table(engine, 'artifact_blob_locations', locations_indices,
                          locations_columns)

    def _pre_upgrade_042(self, engine):
        meta = sqlalchemy.MetaData()
        meta.bind = engine

        metadef_namespaces = sqlalchemy.Table('metadef_namespaces', meta,
                                              autoload=True)
        metadef_objects = sqlalchemy.Table('metadef_objects', meta,
                                           autoload=True)
        metadef_properties = sqlalchemy.Table('metadef_properties', meta,
                                              autoload=True)
        metadef_tags = sqlalchemy.Table('metadef_tags', meta, autoload=True)
        metadef_resource_types = sqlalchemy.Table('metadef_resource_types',
                                                  meta, autoload=True)
        metadef_ns_res_types = sqlalchemy.Table(
            'metadef_namespace_resource_types',
            meta, autoload=True)

        # These will be dropped and recreated as unique constraints.
        self.assertTrue(index_exist('ix_metadef_namespaces_namespace',
                                    metadef_namespaces.name, engine))
        self.assertTrue(index_exist('ix_metadef_objects_namespace_id',
                                    metadef_objects.name, engine))
        self.assertTrue(index_exist('ix_metadef_properties_namespace_id',
                                    metadef_properties.name, engine))
        self.assertTrue(index_exist('ix_metadef_tags_namespace_id',
                                    metadef_tags.name, engine))
        self.assertTrue(index_exist('ix_metadef_resource_types_name',
                                    metadef_resource_types.name, engine))

        # This one will be dropped - not needed
        self.assertTrue(index_exist(
            'ix_metadef_ns_res_types_res_type_id_ns_id',
            metadef_ns_res_types.name, engine))

        # The rest must remain
        self.assertTrue(index_exist('ix_metadef_namespaces_owner',
                                    metadef_namespaces.name, engine))
        self.assertTrue(index_exist('ix_metadef_objects_name',
                                    metadef_objects.name, engine))
        self.assertTrue(index_exist('ix_metadef_properties_name',
                                    metadef_properties.name, engine))
        self.assertTrue(index_exist('ix_metadef_tags_name',
                                    metadef_tags.name, engine))
        self.assertTrue(index_exist('ix_metadef_ns_res_types_namespace_id',
                                    metadef_ns_res_types.name, engine))

        # To be created
        self.assertFalse(unique_constraint_exist
                         ('uq_metadef_objects_namespace_id_name',
                          metadef_objects.name, engine)
                         )
        self.assertFalse(unique_constraint_exist
                         ('uq_metadef_properties_namespace_id_name',
                          metadef_properties.name, engine)
                         )
        self.assertFalse(unique_constraint_exist
                         ('uq_metadef_tags_namespace_id_name',
                          metadef_tags.name, engine)
                         )
        self.assertFalse(unique_constraint_exist
                         ('uq_metadef_namespaces_namespace',
                          metadef_namespaces.name, engine)
                         )
        self.assertFalse(unique_constraint_exist
                         ('uq_metadef_resource_types_name',
                          metadef_resource_types.name, engine)
                         )

    def _check_042(self, engine, data):
        meta = sqlalchemy.MetaData()
        meta.bind = engine

        metadef_namespaces = sqlalchemy.Table('metadef_namespaces', meta,
                                              autoload=True)
        metadef_objects = sqlalchemy.Table('metadef_objects', meta,
                                           autoload=True)
        metadef_properties = sqlalchemy.Table('metadef_properties', meta,
                                              autoload=True)
        metadef_tags = sqlalchemy.Table('metadef_tags', meta, autoload=True)
        metadef_resource_types = sqlalchemy.Table('metadef_resource_types',
                                                  meta, autoload=True)
        metadef_ns_res_types = sqlalchemy.Table(
            'metadef_namespace_resource_types',
            meta, autoload=True)

        # Dropped for unique constraints
        self.assertFalse(index_exist('ix_metadef_namespaces_namespace',
                                     metadef_namespaces.name, engine))
        self.assertFalse(index_exist('ix_metadef_objects_namespace_id',
                                     metadef_objects.name, engine))
        self.assertFalse(index_exist('ix_metadef_properties_namespace_id',
                                     metadef_properties.name, engine))
        self.assertFalse(index_exist('ix_metadef_tags_namespace_id',
                                     metadef_tags.name, engine))
        self.assertFalse(index_exist('ix_metadef_resource_types_name',
                                     metadef_resource_types.name, engine))

        # Dropped - not needed because of the existing primary key
        self.assertFalse(index_exist(
            'ix_metadef_ns_res_types_res_type_id_ns_id',
            metadef_ns_res_types.name, engine))

        # Still exist as before
        self.assertTrue(index_exist('ix_metadef_namespaces_owner',
                                    metadef_namespaces.name, engine))
        self.assertTrue(index_exist('ix_metadef_ns_res_types_namespace_id',
                                    metadef_ns_res_types.name, engine))
        self.assertTrue(index_exist('ix_metadef_objects_name',
                                    metadef_objects.name, engine))
        self.assertTrue(index_exist('ix_metadef_properties_name',
                                    metadef_properties.name, engine))
        self.assertTrue(index_exist('ix_metadef_tags_name',
                                    metadef_tags.name, engine))

        self.assertTrue(unique_constraint_exist
                        ('uq_metadef_namespaces_namespace',
                         metadef_namespaces.name, engine)
                        )
        self.assertTrue(unique_constraint_exist
                        ('uq_metadef_objects_namespace_id_name',
                         metadef_objects.name, engine)
                        )
        self.assertTrue(unique_constraint_exist
                        ('uq_metadef_properties_namespace_id_name',
                         metadef_properties.name, engine)
                        )
        self.assertTrue(unique_constraint_exist
                        ('uq_metadef_tags_namespace_id_name',
                         metadef_tags.name, engine)
                        )
        self.assertTrue(unique_constraint_exist
                        ('uq_metadef_resource_types_name',
                         metadef_resource_types.name, engine)
                        )

    def assert_table(self, engine, table_name, indices, columns):
        table = db_utils.get_table(engine, table_name)
        index_data = [(index.name, index.columns.keys()) for index in
                      table.indexes]
        column_data = [column.name for column in table.columns]
        self.assertItemsEqual(columns, column_data)
        self.assertItemsEqual(indices, index_data)


class TestMigrations(test_base.DbTestCase, test_utils.BaseTestCase):

    def test_no_downgrade(self):
        migrate_file = versions.__path__[0]
        for parent, dirnames, filenames in os.walk(migrate_file):
            for filename in filenames:
                if filename.split('.')[1] == 'py':
                    model_name = filename.split('.')[0]
                    model = __import__(
                        'glance.db.sqlalchemy.migrate_repo.versions.' +
                        model_name)
                    obj = getattr(getattr(getattr(getattr(getattr(
                        model, 'db'), 'sqlalchemy'), 'migrate_repo'),
                        'versions'), model_name)
                    func = getattr(obj, 'downgrade', None)
                    self.assertIsNone(func)


class TestMysqlMigrations(test_base.MySQLOpportunisticTestCase,
                          MigrationsMixin):

    def test_mysql_innodb_tables(self):
        migration.db_sync(engine=self.migrate_engine)

        total = self.migrate_engine.execute(
            "SELECT COUNT(*) "
            "FROM information_schema.TABLES "
            "WHERE TABLE_SCHEMA='%s'"
            % self.migrate_engine.url.database)
        self.assertGreater(total.scalar(), 0, "No tables found. Wrong schema?")

        noninnodb = self.migrate_engine.execute(
            "SELECT count(*) "
            "FROM information_schema.TABLES "
            "WHERE TABLE_SCHEMA='%s' "
            "AND ENGINE!='InnoDB' "
            "AND TABLE_NAME!='migrate_version'"
            % self.migrate_engine.url.database)
        count = noninnodb.scalar()
        self.assertEqual(0, count, "%d non InnoDB tables created" % count)


class TestPostgresqlMigrations(test_base.PostgreSQLOpportunisticTestCase,
                               MigrationsMixin):
    pass


class TestSqliteMigrations(test_base.DbTestCase,
                           MigrationsMixin):
    def test_walk_versions(self):
        # No more downgrades
        self._walk_versions(False, False)


class ModelsMigrationSyncMixin(object):

    def get_metadata(self):
        for table in models_metadef.BASE_DICT.metadata.sorted_tables:
            models.BASE.metadata._add_table(table.name, table.schema, table)
        for table in models_glare.BASE.metadata.sorted_tables:
            models.BASE.metadata._add_table(table.name, table.schema, table)
        return models.BASE.metadata

    def get_engine(self):
        return self.engine

    def db_sync(self, engine):
        migration.db_sync(engine=engine)

    # TODO(akamyshikova): remove this method as soon as comparison with Variant
    # will be implemented in oslo.db or alembic
    def compare_type(self, ctxt, insp_col, meta_col, insp_type, meta_type):
        if isinstance(meta_type, types.Variant):
            meta_orig_type = meta_col.type
            insp_orig_type = insp_col.type
            meta_col.type = meta_type.impl
            insp_col.type = meta_type.impl

            try:
                return self.compare_type(ctxt, insp_col, meta_col, insp_type,
                                         meta_type.impl)
            finally:
                meta_col.type = meta_orig_type
                insp_col.type = insp_orig_type
        else:
            ret = super(ModelsMigrationSyncMixin, self).compare_type(
                ctxt, insp_col, meta_col, insp_type, meta_type)
            if ret is not None:
                return ret
            return ctxt.impl.compare_type(insp_col, meta_col)

    def include_object(self, object_, name, type_, reflected, compare_to):
        if name in ['migrate_version'] and type_ == 'table':
            return False
        return True


class ModelsMigrationsSyncMysql(ModelsMigrationSyncMixin,
                                test_migrations.ModelsMigrationsSync,
                                test_base.MySQLOpportunisticTestCase):
    pass


class ModelsMigrationsSyncPostgres(ModelsMigrationSyncMixin,
                                   test_migrations.ModelsMigrationsSync,
                                   test_base.PostgreSQLOpportunisticTestCase):
    pass


class ModelsMigrationsSyncSQLite(ModelsMigrationSyncMixin,
                                 test_migrations.ModelsMigrationsSync,
                                 test_base.DbTestCase):
    pass



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys
import uuid

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils

from glance.common.glare import definitions
from glance.common import utils
from glance.glare.domain import proxy
from glance.i18n import _LE, _LW

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class ArtifactFactoryProxy(proxy.ArtifactFactory):
    def __init__(self, factory, context, store_api, store_utils):
        self.context = context
        self.store_api = store_api
        self.store_utils = store_utils
        proxy_kwargs = {'store_api': store_api,
                        'store_utils': store_utils,
                        'context': self.context}
        super(ArtifactFactoryProxy, self).__init__(
            factory,
            artifact_proxy_class=ArtifactProxy,
            artifact_proxy_kwargs=proxy_kwargs)


class ArtifactProxy(proxy.Artifact):
    def __init__(self, artifact, context, store_api, store_utils):
        self.artifact = artifact
        self.context = context
        self.store_api = store_api
        self.store_utils = store_utils
        super(ArtifactProxy,
              self).__init__(artifact,
                             proxy_class=ArtifactBlobProxy,
                             proxy_kwargs={"context": self.context,
                                           "store_api": self.store_api})

    def set_type_specific_property(self, prop_name, value):
        if prop_name not in self.artifact.metadata.attributes.blobs:
            super(ArtifactProxy, self).set_type_specific_property(prop_name,
                                                                  value)
            return
        item_key = "%s.%s" % (self.artifact.id, prop_name)
        # XXX FIXME have to add support for BinaryObjectList properties
        blob = definitions.Blob(item_key=item_key)
        blob_proxy = self.helper.proxy(blob)

        if value is None:
            for location in blob_proxy.locations:
                blob_proxy.delete_from_store(location)
        else:
            data = value[0]
            size = value[1]
            blob_proxy.upload_to_store(data, size)
        setattr(self.artifact, prop_name, blob)

    def get_type_specific_property(self, prop_name):
        base = super(ArtifactProxy, self).get_type_specific_property(prop_name)
        if base is None:
            return None
        if prop_name in self.artifact.metadata.attributes.blobs:
            if isinstance(self.artifact.metadata.attributes.blobs[prop_name],
                          list):
                return ArtifactBlobProxyList(self.artifact.id,
                                             prop_name,
                                             base,
                                             self.context,
                                             self.store_api)
            else:
                return self.helper.proxy(base)
        else:
            return base


class ArtifactRepoProxy(proxy.ArtifactRepo):
    def __init__(self, artifact_repo, context, store_api, store_utils):
        self.context = context
        self.store_api = store_api
        proxy_kwargs = {'context': context, 'store_api': store_api,
                        'store_utils': store_utils}
        super(ArtifactRepoProxy, self).__init__(
            artifact_repo,
            proxy_helper=proxy.ArtifactHelper(ArtifactProxy, proxy_kwargs))

    def get(self, *args, **kwargs):
        return self.helper.proxy(self.base.get(*args, **kwargs))


class ArtifactBlobProxy(proxy.ArtifactBlob):
    def __init__(self, blob, context, store_api):
        self.context = context
        self.store_api = store_api
        self.blob = blob
        super(ArtifactBlobProxy, self).__init__(blob)

    def delete_from_store(self, location):
        try:
            ret = self.store_api.delete_from_backend(location['value'],
                                                     context=self.context)
            location['status'] = 'deleted'
            return ret
        except self.store_api.NotFound:
            msg = _LW('Failed to delete blob'
                      ' %s in store from URI') % self.blob.id
            LOG.warn(msg)
        except self.store_api.StoreDeleteNotSupported as e:
            LOG.warn(encodeutils.exception_to_unicode(e))
        except self.store_api.UnsupportedBackend:
            exc_type = sys.exc_info()[0].__name__
            msg = (_LE('Failed to delete blob'
                       ' %(blob_id)s from store: %(exc)s') %
                   dict(blob_id=self.blob.id, exc=exc_type))
            LOG.error(msg)

    def upload_to_store(self, data, size):
        if size is None:  # NOTE(ativelkov): None is "unknown size"
            size = 0
        location, ret_size, checksum, loc_meta = self.store_api.add_to_backend(
            CONF,
            self.blob.item_key,
            utils.LimitingReader(utils.CooperativeReader(data),
                                 CONF.image_size_cap),
            size,
            context=self.context)
        self.blob.size = ret_size
        self.blob.locations = [{'status': 'active', 'value': location}]
        self.blob.checksum = checksum

    @property
    def data_stream(self):
        if len(self.locations) > 0:
            err = None
            try:
                for location in self.locations:
                    data, size = self.store_api.get_from_backend(
                        location['value'],
                        context=self.context)
                    return data
            except Exception as e:
                LOG.warn(_LW('Get blob %(name)s data failed: '
                             '%(err)s.')
                         % {'name': self.blob.item_key,
                            'err': encodeutils.exception_to_unicode(e)})
                err = e

            # tried all locations
            LOG.error(_LE('Glance tried all active locations to get data '
                          'for blob %s '
                          'but all have failed.') % self.blob.item_key)
            raise err


class ArtifactBlobProxyList(proxy.List):
    def __init__(self, artifact_id, prop_name, bloblist, context, store_api):
        self.artifact_id = artifact_id
        self.prop_name = prop_name
        self.context = context
        self.store_api = store_api
        super(ArtifactBlobProxyList,
              self).__init__(bloblist,
                             item_proxy_class=ArtifactBlobProxy,
                             item_proxy_kwargs={'context': context,
                                                'store_api': store_api})

    def insert(self, index, value):
        data = value[0]
        size = value[1]
        item_key = "%s.%s.%s" % (self.artifact_id, self.prop_name,
                                 uuid.uuid4())
        blob = definitions.Blob(item_key=item_key)
        blob_proxy = self.helper.proxy(blob)
        blob_proxy.upload_to_store(data, size)
        super(ArtifactBlobProxyList, self).insert(index, blob_proxy)

    def __setitem__(self, index, value):
        blob = self[index]
        data = value[0]
        size = value[1]
        blob.upload_to_store(data, size)
